import numpy as nm

from sfepy.base.base import Struct
from sfepy.discrete.common.fields import parse_shape, Field
from sfepy.discrete import PolySpace
from sfepy.discrete.fem.mappings import FEMapping
from sfepy.discrete.fem.fields_base import _find_geometry
#from sfepy.discrete.fem.utils import get_min_value
from sfepy.discrete.fem.meshio import convert_complex_output

class L2ConstantVolumeField(Field):
    """
    The L2 constant-in-a-region approximation.
    """
    family_name = 'volume_L2_constant'

    def __init__(self, name, dtype, shape, region, approx_order=0):
        """
        Create a L2 constant field.

        Parameters
        ----------
        name : str
            The field name.
        dtype : numpy.dtype
            The field data type: float64 or complex128.
        shape : int/tuple/str
            The field shape: 1 or (1,) or 'scalar', space dimension (2, or (2,)
            or 3 or (3,)) or 'vector', or a tuple. The field shape determines
            the shape of the FE base functions and is related to the number of
            components of variables and to the DOF per node count, depending
            on the field kind.
        region : Region
            The region where the field is defined.
        approx_order : int or tuple
            The FE approximation order. Ignored here.
        """
        field_dim = region.field_dim if hasattr(region, 'field_dim')\
            else region.domain.shape.dim
        shape = parse_shape(shape, field_dim)
        Struct.__init__(self, name=name, dtype=dtype, shape=shape,
                        region=region, approx_order=0)
        self.domain = self.region.domain
        self.cmesh = self.region.cmesh

        self.gel, self.is_surface = _find_geometry(self.region)
        self._setup_kind()
        self._create_interpolant()

        cells = self.region.get_cells(true_cells_only=False)
        self.econn = nm.zeros((len(cells), 1), dtype=nm.int32)

        self.domain = self.region.domain

        self.n_components = nm.prod(self.shape)
        self.val_shape = self.shape
        self.n_nod = 1

        self.extra_data = {}
        self.mappings = {}

    def _create_interpolant(self):
        name = '%s_%s_%s_%d' % (self.gel.name, self.space,
                                self.poly_space_base, self.approx_order)
        ps = PolySpace.any_from_args(name, self.gel, self.approx_order,
                                     base='lagrange')
        self.poly_space = ps

    def setup_extra_data(self, info):
        pass

    def get_coor(self, nods=None):
        """
        Returns the barycenter of the field region.

        Parameters
        ----------
        nods : array, optional
            Ignored.
        """
        coors = nm.sum(self.domain.mesh.coors[self.region.vertices],
                       axis=0, keepdims=True)
        coors /= self.region.shape.n_vertex
        return coors

    def get_econn(self, conn_type, region, trace_region=None, local=False):
        """
        Get extended connectivity of the given type in the given region.

        Parameters
        ----------
        conn_type: tuple or string
            DOF connectivity type, ignored.
        region: sfepy.discrete.common.region.Region
            The region for which the connectivity is required.
        trace_region: None or string
            Ignored.
        local: bool
            Ignored.

        Returns
        -------
        econn: numpy.ndarray
            The extended connectivity array.
        """
        if region.name == self.region.name:
            conn = self.econn

        else:
            cells = region.get_cells(true_cells_only=False)
            conn = nm.zeros((len(cells), 1), dtype=nm.int32)

        return conn

    def get_data_shape(self, integral, integration='cell', region_name=None):
        """
        Get element data dimensions.

        Parameters
        ----------
        integral : Integral instance
            The integral describing used numerical quadrature.
        integration : 'cell'
            The term integration type. Ignored.
        region_name : str
            The name of the region of the integral.

        Returns
        -------
        data_shape : 4 ints
            The `(n_el, n_qp, dim, n_en)` for volume shape kind,
            `(n_fa, n_qp, dim, n_fn)` for surface shape kind.

        Notes
        -----
        - `n_el`, `n_fa` = number of elements/facets
        - `n_qp` = number of quadrature points per element/facet
        - `dim` = spatial dimension
        - `n_en`, `n_fn` = number of element/facet nodes
        - `n_nod` = number of element nodes
        """
        if integration not in ('cell', 'facet', 'facet_extra'):
            raise NotImplementedError('unsupported integration type! (%s)'
                                      % integration)

        region = self.domain.regions[region_name]
        n_cell = len(region.entities[region.kind_tdim])
        _, weights = integral.get_qp(self.gel.name)
        n_qp = weights.shape[0]
        dim = region.field_dim if hasattr(region, 'field_dim') else region.dim
        return (n_cell, n_qp, dim, 1)

    def get_dofs_in_region(self, region, merge=True):
        """
        Return indices of DOFs that belong to the given region.
        """
        return nm.zeros(1, dtype=nm.int32)

    def get_base(self, key, derivative, integral, iels=None,
                 from_geometry=False, base_only=True):
        qp_coors, qp_weights = integral.get_qp(self.gel.name)
        ps = self.poly_space
        bf = ps.eval_base(qp_coors)
        if key[0] == 'b': # BQP
            num = 6 if self.gel.n_vertex == 4 else 4
            bf = nm.tile(bf, (num, 1, 1, 1))

        if base_only:
            return bf

        else:
            return bf, qp_weights

    def create_mapping(self, region, integral, integration,
                       return_mapping=True):
        domain = self.domain
        coors = domain.get_mesh_coors(actual=True)
        iels = region.get_cells(true_cells_only=(region.kind == 'cell'))

        if integration == 'cell':
            ps = self.poly_space
            geo_ps = self.gel.poly_space
            dconn = domain.get_conn(tdim=region.tdim, cells=iels)
            qp_coors, qp_weights = integral.get_qp(self.gel.name)
            bf = ps.eval_base(qp_coors)

        elif integration in ('facet', 'facet_extra'):
            if self.is_surface:
                gel = self.gel
                ps = self.poly_space

            else:
                gel = self.gel.surface_facet
                ps = PolySpace.any_from_args('aux', gel, self.approx_order,
                                             base='lagrange')

            geo_ps = gel.poly_space
            domain.create_surface_group(region)
            sd = domain.surface_groups[region.name]
            dconn = sd.get_connectivity()
            qp_coors, qp_weights = integral.get_qp(gel.name)
            bf = ps.eval_base(qp_coors)

        mapping = FEMapping(coors, dconn, poly_space=geo_ps)
        out = mapping.get_mapping(qp_coors, qp_weights, bf, poly_space=ps,
                                  is_face=self.is_surface)

        if return_mapping:
            out = (out, mapping)

        return out

    def create_output(self, dofs, var_name, dof_names=None,
                      key=None, extend=True, fill_value=None,
                      linearization=None):
        _new_dofs = nm.tile(dofs, (self.region.shape.n_vertex, 1))
        if extend:
            if fill_value is None:
                fill_value = 0.0

            n_nod = self.domain.shape.n_nod
            new_dofs = nm.full((n_nod, dofs.shape[1]), fill_value,
                               dtype=self.dtype)
            new_dofs[self.region.vertices] = _new_dofs

        else:
            new_dofs = _new_dofs

        out = {}
        out[key] = Struct(name='output_data', mode='vertex',
                          data=new_dofs, var_name=var_name,
                          dofs=dof_names, region_name=self.region.name)

        out = convert_complex_output(out)
        return out

class L2ConstantSurfaceField(L2ConstantVolumeField):
    family_name = 'surface_L2_constant'



import sympy as sm
import numpy as nm
import scipy.sparse as sps

from sfepy.base.base import dict_to_struct
import sfepy.base.testing as tst

conf = {
    'name' : 'semismooth_newton',
    'kind' : 'nls.semismooth_newton',

    'semismooth' : True,

    'i_max'      : 10,
    'eps_a'      : 1e-8,
    'eps_r'      : 1e-2,
    'macheps'   : 1e-16,
    'lin_red'    : 1e-2, # Linear system error < (eps_a * lin_red).
    'ls_red'     : {'regular' : 0.1, 'steepest_descent' : 0.01},
    'ls_red_warp' : 0.001,
    'ls_on'      : 2.0,
    'ls_min'     : 1e-10,
    ## 'log'        : {'plot' : 'aux.png'},
}

ls_conf = {
    'name' : 'ls',
    'kind' : 'ls.scipy_direct',
}

alpha = 30.0 * nm.pi / 180.0
ca = nm.cos(alpha)
sa = nm.sin(alpha)

# Coefficient of friction.
fc = 0.01

# Spring (bar) stiffnesses.
ks = nm.ones((7,), dtype=nm.float64)

# Load vector.
fs = nm.array([0, 0, 1, 1, -1, -1], dtype=nm.float64)


def eval_matrix(mtx, **kwargs):

    num_mtx = nm.zeros(mtx.shape, dtype=nm.float64)

    mtx = mtx.subs(kwargs)
    for ir in range(mtx.shape[0]):
        for ic in range(mtx.shape[1]):
            num_mtx[ir,ic] = float(mtx[ir,ic])

    return num_mtx

def convert_to_csr(m_in):
    m_out = {}
    for key, mtx in m_in.items():
        m_out[key] = sps.csr_matrix(mtx)

    return m_out

def define_matrices():
    e = sm.zeros(7, 1)
    e[0] = sm.sympify('u1')
    e[1] = sm.sympify('c * u3 + s * v3')
    e[2] = sm.sympify('u2 - u1')
    e[3] = sm.sympify('v4 + u2n')
    e[4] = sm.sympify('u4 - u3')
    e[5] = sm.sympify('v3 + u1n')
    e[6] = sm.sympify('c * u4 + s * v4 - c * u1 + s * u1n')
    k = sm.Matrix(7, 1, lambda i, j: sm.Symbol('k%d' % i))

    psi = (k.T * e.applyfunc(lambda x: x**2))[0]
    work = sm.sympify('u1 * g1 + u2 * g2')

    w = ['u1', 'u2', 'u3', 'u4', 'v3', 'v4']
    wbar = w[:2]
    g = ['g1', 'g2']

    mtx_a = sm.zeros(6, 6)
    for ir in range(mtx_a.shape[0]):
        for ic in range(mtx_a.shape[1]):
            mtx_a[ir,ic] = sm.diff(sm.diff(psi, w[ir]), w[ic])

    mtx_b_bar = sm.zeros(2, 6)
    for ir in range(mtx_b_bar.shape[0]):
        for ic in range(mtx_b_bar.shape[1]):
            mtx_b_bar[ir,ic] = sm.diff(sm.diff(work, g[ir]), w[ic])

    mtx_b = sm.zeros(2, 2)
    for ir in range(mtx_b.shape[0]):
        for ic in range(mtx_b.shape[1]):
            mtx_b[ir,ic] = sm.diff(sm.diff(work, wbar[ir]), g[ic])

    mtx_c = sm.eye(2)

    sn1 = sm.diff(psi, 'u1n').subs('u1n', 0)
    sn2 = sm.diff(psi, 'u2n').subs('u2n', 0)
    sn = sm.Matrix([sn1, sn2])

    mtx_d = sm.zeros(2, 6)
    for ir in range(mtx_d.shape[0]):
        for ic in range(mtx_d.shape[1]):
            mtx_d[ir,ic] = sm.diff(sn[ir], w[ic])

    subsd = {'c' : ca, 's' : sa}
    for ii, _k in enumerate(k):
        subsd['k%d' % ii] = ks[ii]

    m = {
        'A' : eval_matrix(mtx_a, **subsd),
        'Bb' : eval_matrix(mtx_b_bar),
        'B' : eval_matrix(mtx_b),
        'C' : eval_matrix(mtx_c),
    }

    ms = convert_to_csr(m)

    m['sn'] = sn.subs(subsd)
    m['D'] = mtx_d.subs(subsd)
    m['fs'] = fs

    return m, ms, w

def test_semismooth_newton():
    import numpy as nm
    from sfepy.solvers import Solver

    m, ms, w_names = define_matrices()

    ns = [0, 6, 2, 2]

    offsets = nm.cumsum(ns)
    nx = offsets[-1]

    iw = slice(offsets[0], offsets[1])
    ig = slice(offsets[1], offsets[2])
    il = slice(offsets[2], offsets[3])

    def fun_smooth(vec_x):
        xw = vec_x[iw]
        xg = vec_x[ig]
        xl = vec_x[il]

        rw = ms['A'] * xw - ms['Bb'].T * xg - m['fs']
        rg = ms['Bb'] * xw + xl * xg

        rwg = nm.r_[rw, rg]

        return rwg

    def fun_smooth_grad(vec_x):
        xl = vec_x[il]
        xg = vec_x[ig]

        mzl = nm.zeros((6, 2), dtype=nm.float64)

        mw = nm.c_[m['A'], -m['Bb'].T, mzl]
        mg = nm.c_[m['Bb'], nm.diag(xl), nm.diag(xg)]

        mx = nm.r_[mw, mg]

        mx = sps.csr_matrix(mx)
        return mx

    def fun_a(vec_x):
        xw = vec_x[iw]
        xg = vec_x[ig]

        subsd = {}
        for ii, key in enumerate(w_names):
            subsd[key] = xw[ii]

        sn = eval_matrix(m['sn'], **subsd).squeeze()

        ra = nm.abs(xg) - fc * nm.abs(sn)

        return -ra

    def fun_a_grad(vec_x):
        xw = vec_x[iw]
        xg = vec_x[ig]
        xl = vec_x[il]

        subsd = {}
        for ii, key in enumerate(w_names):
            subsd[key] = xw[ii]

        md = eval_matrix(m['D'], **subsd)
        sn = eval_matrix(m['sn'], **subsd).squeeze()

        ma = nm.zeros((xl.shape[0], nx), dtype=nm.float64)
        ma[:,iw] = - fc * nm.sign(sn)[:,None] * md
        ma[:,ig] = nm.sign(xg)[:,None] * m['C']

        return -sps.csr_matrix(ma)

    def fun_b(vec_x):
        xl = vec_x[il]

        return xl

    def fun_b_grad(vec_x):
        xl = vec_x[il]

        mb = nm.zeros((xl.shape[0], nx), dtype=nm.float64)
        mb[:,il] = m['C']

        return sps.csr_matrix(mb)

    vec_x0 = 0.1 * nm.ones((nx,), dtype=nm.float64)

    lin_solver = Solver.any_from_conf(dict_to_struct(ls_conf))
    status = {}
    solver = Solver.any_from_conf(dict_to_struct(conf),
                                  fun_smooth=fun_smooth,
                                  fun_smooth_grad=fun_smooth_grad,
                                  fun_a=fun_a,
                                  fun_a_grad=fun_a_grad,
                                  fun_b=fun_b,
                                  fun_b_grad=fun_b_grad,
                                  lin_solver=lin_solver,
                                  status=status)

    vec_x = solver(vec_x0)

    xw = vec_x[iw]
    xg = vec_x[ig]
    xl = vec_x[il]

    tst.report('x:', xw)
    tst.report('g:', xg)
    tst.report('l:', xl)

    subsd = {}
    for ii, key in enumerate(w_names):
        subsd[key] = xw[ii]

    sn = eval_matrix(m['sn'], **subsd).squeeze()
    tst.report('sn:', sn)

    assert status['condition'] == 0



r"""
The linear elastodynamics of an elastic body loaded by a given base motion.

Find :math:`\ul{u}` such that:

.. math::
    \int_{\Omega} \rho \ul{v} \pddiff{\ul{u}}{t}
    + \int_{\Omega} D_{ijkl}\ e_{ij}(\ul{v}) e_{kl}(\ul{u})
    = 0
    \;, \quad \forall \ul{v} \;, \\
    u_1(t) =  10^{-5} \sin(\omega t) \sin(k x_2)
    \mbox{ on } \Gamma_\mathrm{Seismic} \;, \\
    \omega = c_L k \;,

where :math:`c_L` is the longitudinal wave propagation speed, :math:`k = 2 \pi
/ L`,` :math:`L` is the length of the domain and

.. math::
    D_{ijkl} = \mu (\delta_{ik} \delta_{jl}+\delta_{il} \delta_{jk}) +
    \lambda \ \delta_{ij} \delta_{kl}
    \;.

See :ref:`linear_elasticity-elastodynamic` example for notes on elastodynamics
solvers.

Usage Examples
--------------

Run with the default settings (the Newmark method, 2D problem, results stored
in ``output/seismic/``)::

  sfepy-run sfepy/examples/linear_elasticity/seismic_load.py -o tsn

View the resulting displacements on the deforming mesh (10x magnified)::

  sfepy-view output/seismic/tsn.h5 -2 -f u:wu:f10:p0 1:vw:p0

Use the central difference explicit method with the reciprocal mass matrix
algorithm [1]_ and view the resulting stress waves::

  sfepy-run sfepy/examples/linear_elasticity/seismic_load.py  -d "dims=(5e-3, 5e-3), shape=(51, 51), tss_name=tscd, ls_name=lsrmm, mass_beta=0.5, mass_lumping=row_sum, fast_rmm=True, save_times=all" -o tscd

  sfepy-view output/seismic/tscd.h5 -2 -f cauchy_stress:wu:f10:p0 1:vw:p0

.. [1] González, J.A., Kolman, R., Cho, S.S., Felippa, C.A., Park, K.C., 2018.
       Inverse mass matrix via the method of localized Lagrange multipliers.
       International Journal for Numerical Methods in Engineering 113, 277–295.
       https://doi.org/10.1002/nme.5613
"""
import numpy as nm

import sfepy.mechanics.matcoefs as mc
from sfepy.discrete.fem.meshio import UserMeshIO
from sfepy.mesh.mesh_generators import gen_block_mesh

def define(
        E=200e9, nu=0.3, rho=7800,
        plane='strain',
        dims=(5e-3, 5e-3),
        shape=(31, 31),
        v0=1.0,
        ct1=1.5,
        dt=None,
        edt_safety=0.2,
        tss_name='tsn',
        tsc_name='tscedl',
        adaptive=False,
        ls_name='lsd',
        mass_beta=0.0,
        mass_lumping='none',
        fast_rmm=False,
        active_only=False,
        save_times=20,
        output_dir='output/seismic',
):
    """
    Parameters
    ----------
    E, nu, rho: material parameters
    plane: plane strain or stress hypothesis
    dims: physical dimensions of the block (L, d, x)
    shape: numbers of mesh vertices along each axis
    v0: initial impact velocity
    ct1: final time in L / "longitudinal wave speed" units
    dt: time step (None means automatic)
    edt_safety: safety factor time step multiplier for explicit schemes,
                if dt is None
    tss_name: time stepping solver name (see "solvers" section)
    tsc_name: time step controller name (see "solvers" section)
    adaptive: use adaptive time step control
    ls_name: linear system solver name (see "solvers" section)
    mass_beta: averaged mass matrix parameter 0 <= beta <= 1
    mass_lumping: mass matrix lumping ('row_sum', 'hrz' or 'none')
    fast_rmm: use zero inertia term with lsrmm
    save_times: number of solutions to save
    output_dir: output directory
    """
    dim = len(dims)

    lam, mu = mc.lame_from_youngpoisson(E, nu, plane=plane)
    # Longitudinal and shear wave propagation speeds.
    cl = nm.sqrt((lam + 2.0 * mu) / rho)
    cs = nm.sqrt(mu / rho)

    # Element size.
    L, d = dims[:2]
    H = L / (nm.max(shape) - 1)

    # Time-stepping parameters.
    if dt is None:
        # For implicit schemes, dt based on the Courant number C0 = dt * cl / H
        # equal to 1.
        dt = H / cl # C0 = 1
        if tss_name in ('tsvv', 'tscd'):
            # For explicit schemes, use a safety margin.
            dt *= edt_safety

    t1 = ct1 * L / cl

    def mesh_hook(mesh, mode):
        """
        Generate the block mesh.
        """
        if mode == 'read':
            mesh = gen_block_mesh(dims, shape, 0.5 * nm.array(dims),
                                  name='user_block', verbose=False)
            return mesh

        elif mode == 'write':
            pass

    def post_process(out, problem, state, extend=False):
        """
        Calculate and output strain and stress for given displacements.
        """
        from sfepy.base.base import Struct

        ev = problem.evaluate
        strain = ev('ev_cauchy_strain.i.Omega(u)', mode='el_avg', verbose=False)
        stress = ev('ev_cauchy_stress.i.Omega(solid.D, u)', mode='el_avg',
                    copy_materials=False, verbose=False)

        out['cauchy_strain'] = Struct(name='output_data', mode='cell',
                                      data=strain)
        out['cauchy_stress'] = Struct(name='output_data', mode='cell',
                                      data=stress)

        return out

    filename_mesh = UserMeshIO(mesh_hook)

    regions = {
        'Omega' : 'all',
        'Seismic' : ('vertices in (x < 1e-12)', 'facet'),
    }

    # Iron.
    materials = {
        'solid' : ({
            'D': mc.stiffness_from_youngpoisson(dim=dim, young=E, poisson=nu,
                                                plane=plane),
            'rho': rho,
            '.lumping' : mass_lumping,
            '.beta' : mass_beta,
        },),
    }

    fields = {
        'displacement': ('real', 'vector', 'Omega', 1),
    }

    integrals = {
        'i' : 2,
    }

    variables = {
        'u' : ('unknown field', 'displacement', 0),
        'v' : ('test field', 'displacement', 'u'),
    }

    def get_ebcs(ts, coors, mode='u'):
        y = coors[:, 1]
        amplitude = 0.00001
        k = 2 * nm.pi / L
        omega = cl * k
        if mode == 'u':
            val = amplitude * nm.sin(ts.time * omega) * nm.sin(k * y)

        elif mode == 'du':
            val = amplitude * omega * nm.cos(ts.time * omega) * nm.sin(k * y)

        elif mode == 'ddu':
            val = -amplitude * omega**2 * nm.sin(ts.time * omega) * nm.sin(k * y)

        return val

    functions = {
        'get_u' : (lambda ts, coor, **kwargs: get_ebcs(ts, coor),),
        'get_du' : (lambda ts, coor, **kwargs: get_ebcs(ts, coor, mode='du'),),
        'get_ddu' : (lambda ts, coor, **kwargs: get_ebcs(ts, coor, mode='ddu'),),
    }

    ebcs = {
        'Seismic' : ('Seismic', {'u.0' : 'get_u', 'du.0' : 'get_du',
                                 'ddu.0' : 'get_ddu'}),
    }

    ics = {
        'ic' : ('Omega', {'u.all' : 0.0, 'du.all' : 0.0}),
    }

    if (ls_name == 'lsrmm') and fast_rmm:
        # Speed up residual calculation, as M is not used with lsrmm.
        term = 'dw_zero.i.Omega(v, ddu)'

    else:
        term = 'de_mass.i.Omega(solid.rho, solid.lumping, solid.beta, v, ddu)'

    equations = {
        'balance_of_forces' :
        term + '+ dw_lin_elastic.i.Omega(solid.D, v, u) = 0',
    }

    solvers = {
        'lsd' : ('ls.auto_direct', {
            # Reuse the factorized linear system from the first time step.
            'use_presolve' : True,
            # Speed up the above by omitting the matrix digest check used
            # normally for verification that the current matrix corresponds to
            # the factorized matrix stored in the solver instance. Use with
            # care!
            'use_mtx_digest' : False,
        }),
        'lsi' : ('ls.petsc', {
            'method' : 'cg',
            'precond' : 'icc',
            'i_max' : 150,
            'eps_a' : 1e-32,
            'eps_r' : 1e-8,
            'verbose' : 2,
        }),
        'lsrmm' : ('ls.rmm', {
            'rmm_term' : """de_mass.i.Omega(solid.rho, solid.lumping,
                                            solid.beta, v, ddu)""",
            'debug' : False,
        }),
        'newton' : ('nls.newton', {
            'i_max'      : 1,
            'eps_a'      : 1e-6,
            'eps_r'      : 1e-6,
            'ls_on'      : 1e100,
        }),
        'tsvv' : ('ts.velocity_verlet', {
            # Explicit method.
            't0' : 0.0,
            't1' : t1,
            'dt' : dt,
            'n_step' : None,

            'is_linear'  : True,

            'verbose' : 1,
        }),
        'tscd' : ('ts.central_difference', {
            # Explicit method. Supports ls.rmm.
            't0' : 0.0,
            't1' : t1,
            'dt' : dt,
            'n_step' : None,

            'is_linear'  : True,

            'verbose' : 1,
        }),
        'tsn' : ('ts.newmark', {
            't0' : 0.0,
            't1' : t1,
            'dt' : dt,
            'n_step' : None,

            'is_linear'  : True,

            'beta' : 0.25,
            'gamma' : 0.5,

            'verbose' : 1,
        }),
        'tsga' : ('ts.generalized_alpha', {
            't0' : 0.0,
            't1' : t1,
            'dt' : dt,
            'n_step' : None,

            'is_linear'  : True,

            'rho_inf' : 0.5,
            'alpha_m' : None,
            'alpha_f' : None,
            'beta' : None,
            'gamma' : None,

            'verbose' : 1,
        }),
        'tsb' : ('ts.bathe', {
            't0' : 0.0,
            't1' : t1,
            'dt' : dt,
            'n_step' : None,

            'is_linear'  : True,

            'verbose' : 1,
        }),
        'tscedb' : ('tsc.ed_basic', {
            'eps_r' : (1e-4, 1e-1),
            'eps_a' : (1e-8, 5e-2),
            'fmin' : 0.3,
            'fmax' : 2.5,
            'fsafety' : 0.85,
        }),
        'tscedl' : ('tsc.ed_linear', {
            'eps_r' : (1e-4, 1e-1),
            'eps_a' : (1e-8, 5e-2),
            'fmin' : 0.3,
            'fmax' : 2.5,
            'fsafety' : 0.85,
            'red_factor' : 0.9,
            'inc_wait' : 10,
            'min_inc_factor' : 1.5,
        }),
    }

    options = {
        'ts' : tss_name,
        'tsc' : tsc_name if adaptive else None,
        'nls' : 'newton',
        'ls' : ls_name,

        'save_times' : save_times,

        'active_only' : active_only,
        'auto_transform_equations' : True,

        'output_format' : 'h5',
        'output_dir' : output_dir,
        'post_process_hook' : 'post_process',
    }

    return locals()



import os.path as op
import numpy as nm
import pytest

import sfepy.base.testing as tst

def fix_u_fun(ts, coors, bc=None, problem=None, extra_arg=None):
    return nm.zeros_like(coors)

@pytest.fixture(scope='module')
def data():
    import sfepy
    from sfepy.base.base import Struct
    from sfepy.discrete.fem import Mesh, FEDomain, Field
    mesh = Mesh.from_file('meshes/2d/rectangle_tri.mesh',
                          prefix_dir=sfepy.data_dir)
    domain = FEDomain('domain', mesh)
    dim = domain.shape.dim

    min_x, max_x = domain.get_mesh_bounding_box()[:,0]
    eps = 1e-8 * (max_x - min_x)

    omega = domain.create_region('Omega', 'all')
    gamma1 = domain.create_region('Gamma1',
                                  'vertices in x < %.10f' % (min_x + eps),
                                  'facet')
    gamma2 = domain.create_region('Gamma2',
                                  'vertices in x > %.10f' % (max_x - eps),
                                  'facet')

    field = Field.from_args('fu', nm.float64, 'vector', omega,
                            approx_order=2)

    return Struct(dim=dim, omega=omega, gamma1=gamma1, gamma2=gamma2,
                  field=field)

def test_term_evaluation(data):
    from sfepy.discrete import Integral, FieldVariable
    from sfepy.terms.terms import Term

    integral = Integral('i', order=3)

    u = FieldVariable('u', 'parameter', data.field,
                      primary_var_name='(set-to-None)')

    term = Term.new('ev_volume(u)', integral, data.omega, u=u)
    term *= 10.0

    term.setup()

    vol = term.evaluate()

    tst.report('volume: %.8f == 2000.0' % vol)
    assert nm.allclose(vol, 2000.0, rtol=1e-15, atol=0)

def test_term_arithmetics(data):
    from sfepy.discrete import FieldVariable, Integral
    from sfepy.terms.terms import Term

    integral = Integral('i', order=3)

    u = FieldVariable('u', 'parameter', data.field,
                      primary_var_name='(set-to-None)')

    t1 = Term.new('ev_volume(u)', integral, data.omega, u=u)
    t2 = Term.new('ev_volume(u)', integral, data.gamma1, u=u)

    expr = 2.2j * (t1 * 5.5 - 3j * t2) * 0.25

    ok = len(expr) == 2
    if not ok:
        tst.report('wrong expression length!')

    _ok = nm.allclose(expr[0].sign, 3.025j, rtol=1e-15, atol=0)
    if not _ok:
        tst.report('wrong sign of the first term!')
    ok = ok and _ok

    _ok = nm.allclose(expr[1].sign, 1.65, rtol=1e-15, atol=0)
    if not _ok:
        tst.report('wrong sign of the second term!')
    ok = ok and _ok

    assert ok

def test_variables(data):
    from sfepy.discrete import FieldVariable, Integral

    ok = True

    u = FieldVariable('u', 'parameter', data.field,
                      primary_var_name='(set-to-None)')
    u.set_constant(1.0)
    vec = u() # Nodal values.

    _ok = nm.allclose(vec, 1.0)
    tst.report('set constant:', _ok)
    ok = _ok and ok

    def fun(coors):
        val = nm.empty_like(coors)
        val[:, 0] = 2 * coors[:, 1] - coors[:, 0]
        val[:, 1] = coors[:, 0] + 3 * coors[:, 1]
        return val
    u.set_from_function(fun)

    coors = u.field.get_coor()
    eu = u.evaluate_at(coors)[..., 0]

    _ok = nm.allclose(eu, fun(coors), rtol=0.0, atol=1e-13)
    tst.report('set from function:', _ok)
    ok = _ok and ok

    integral = Integral('i', order=2)
    gu_qp = u.evaluate(mode='grad', integral=integral)

    # du_i/dx_j, i = column, j = row.
    gu = nm.array([[-1.,  1.],
                   [ 2.,  3.]])
    _ok = nm.allclose(gu_qp, gu[None, None, ...], rtol=0.0, atol=1e-13)
    tst.report('set from function - gradient:', _ok)
    ok = _ok and ok

    u_qp = gu_qp[..., :, :1]
    u.set_from_qp(u_qp, integral)
    vu = u()

    _ok = (nm.allclose(vu[::2], -1, rtol=0.0, atol=1e-13) and
           nm.allclose(vu[1::2], 2, rtol=0.0, atol=1e-13))
    tst.report('set from qp:', _ok)
    ok = _ok and ok

    assert ok

def test_solving(data, output_dir):
    from sfepy.base.base import IndexedStruct
    from sfepy.discrete import (FieldVariable, Material, Problem, Function,
                                Equation, Equations, Integral)
    from sfepy.discrete.conditions import Conditions, EssentialBC
    from sfepy.terms import Term
    from sfepy.solvers.ls import ScipyDirect
    from sfepy.solvers.nls import Newton
    from sfepy.mechanics.matcoefs import stiffness_from_lame

    u = FieldVariable('u', 'unknown', data.field)
    v = FieldVariable('v', 'test', data.field, primary_var_name='u')

    m = Material('m', D=stiffness_from_lame(data.dim, 1.0, 1.0))
    f = Material('f', val=[[0.02], [0.01]])

    bc_fun = Function('fix_u_fun', fix_u_fun,
                      extra_args={'extra_arg' : 'hello'})

    fix_u = EssentialBC('fix_u', data.gamma1, {'u.all' : bc_fun})
    shift_u = EssentialBC('shift_u', data.gamma2, {'u.0' : 0.1})

    integral = Integral('i', order=3)

    t1 = Term.new('dw_lin_elastic(m.D, v, u)',
                  integral, data.omega, m=m, v=v, u=u)

    t2 = Term.new('dw_volume_lvf(f.val, v)', integral, data.omega, f=f, v=v)

    eq = Equation('balance', t1 + t2)
    eqs = Equations([eq])

    ls = ScipyDirect({})

    nls_status = IndexedStruct()
    nls = Newton({}, lin_solver=ls, status=nls_status)

    pb = Problem('elasticity', equations=eqs)
    ## pb.save_regions_as_groups('regions')

    pb.set_bcs(ebcs=Conditions([fix_u, shift_u]))
    pb.set_solver(nls)

    state = pb.solve(save_results=False)

    name = op.join(output_dir, 'test_high_level_solving.vtk')
    pb.save_state(name, state)

    ok = nls_status.condition == 0
    if not ok:
        tst.report('solver did not converge!')

    _ok = state.has_ebc()
    if not _ok:
        tst.report('EBCs violated!')

    ok = ok and _ok

    assert ok



import numpy as nm
from sfepy import data_dir
from sfepy.mesh.splinebox import SplineBox, SplineRegion2D
from sfepy.mesh.bspline import BSpline
from sfepy.discrete.fem import Mesh
import sfepy.base.testing as tst

def tetravolume(cells, vertices):
    vol = 0.0
    c1 = nm.ones((4,4), dtype=nm.float64)
    mul = 1.0 / 6.0
    for ic in cells:
        c1[:,:3] = vertices[ic,:]
        vol += mul * nm.linalg.det(c1)

    return -vol

tolerance = 1e-6

def test_spbox_3d():
    """
    Check volume change of the mesh which is deformed using
    the SplineBox functions.
    """
    from sfepy.discrete.fem import Mesh
    from sfepy.mesh.splinebox import SplineBox
    mesh = Mesh.from_file(data_dir + '/meshes/3d/cylinder.vtk')
    conn = mesh.get_conn('3_4')
    vol0 = tetravolume(conn, mesh.coors)

    bbox = nm.array(mesh.get_bounding_box()).T
    spbox = SplineBox(bbox, mesh.coors)
    cpoints0 = spbox.get_control_points(init=True)

    for ii in range(4):
        for jj in range(4):
            spbox.move_control_point((0, ii, jj), [-0.02, 0, 0])
    coors = spbox.evaluate()
    vol1 = tetravolume(conn, coors)
    mesh.coors[:] = coors

    spbox.set_control_points(cpoints0)
    coors = spbox.evaluate()
    vol2 = tetravolume(conn, coors)

    ok = True
    actual_volumes = (vol0, vol1, vol2)
    expected_volumes = (1.22460186e-4, 1.46950423e-4, 1.22460186e-4)

    for ii in range(3):
        relerr = abs(actual_volumes[ii] - expected_volumes[ii])\
                 / expected_volumes[ii]
        ok = ok and (relerr < tolerance)

    if not ok:
        tst.report('expected volumes:')
        tst.report(expected_volumes)
        tst.report('actual volumes:')
        tst.report(actual_volumes)

    assert ok

def test_spbox_2d():
    """
    Check position of a given vertex in the deformed mesh.
    """
    mesh = Mesh.from_file(data_dir + '/meshes/2d/square_tri1.mesh')
    spb = SplineBox([[-1, 1], [-1, 0.6]], mesh.coors, nsg=[2,1])
    spb.move_control_point(1, [0.1, -0.2])
    spb.move_control_point(2, [0.2, -0.3])
    spb.move_control_point(3, [0.0, -0.1])

    pt0 = mesh.coors[175,:].copy()
    mesh.cmesh.coors[:] = spb.evaluate()
    pt1 = mesh.coors[175,:]

    expected_distance = 0.165892726387
    actual_distance = nm.linalg.norm(pt0 - pt1)
    ok = nm.fabs(actual_distance - expected_distance)\
        / expected_distance < tolerance

    if not ok:
        tst.report('expected distance:')
        tst.report(expected_distance)
        tst.report('actual distance:')
        tst.report(actual_distance)

    assert ok

def test_spbox_field():
    """
    'Field' vs. 'coors'.
    """
    mesh = Mesh.from_file(data_dir + '/meshes/2d/its2D.mesh')
    coors = mesh.coors.copy()
    bbox = nm.vstack((nm.amin(coors, 0), nm.amax(coors, 0))).T

    coors_1 = coors.copy()

    alpha = coors[:, 0]
    spbox = SplineBox(bbox, coors, nsg=[1, 2], field=alpha)
    dv1 = spbox.evaluate_derivative(6, 1)
    spbox.move_control_point(6, -0.2)
    c1 = spbox.evaluate()
    coors_1[:, 0] = c1[:, 0]

    alpha = coors[:, 1]
    spbox = SplineBox(bbox, coors, nsg=[1, 2], field=alpha)
    dv2 = spbox.evaluate_derivative(6, 1)
    spbox.move_control_point(6, 0.2)
    c2 = spbox.evaluate()
    coors_1[:, 1] = c2[:, 0]

    spbox = SplineBox(bbox, coors, nsg=[1, 2])
    dv = spbox.evaluate_derivative(6, [1, 1])
    spbox.move_control_point(6, [-0.2, 0.2])
    coors_2 = spbox.evaluate()

    rel_coor_dist = nm.linalg.norm(coors_2 - coors_1)\
        / nm.linalg.norm(coors_2)
    ok = rel_coor_dist < tolerance
    rel_dvel_dist = nm.linalg.norm(dv - nm.hstack([dv1, dv2]))\
        / nm.linalg.norm(dv)
    ok = ok and rel_dvel_dist < tolerance

    if not ok:
        tst.report('modified coordinates do not match, relative error:')
        tst.report(rel_coor_dist)
        tst.report('derivatives do not match, relative error:')
        tst.report(rel_dvel_dist)

    assert ok

def test_spregion2d():
    """
    Check position of a given vertex in the deformed mesh.
    """
    line_l = nm.array([[-1, 1], [-1, .5], [-1, 0], [-1, -.5]])
    line_r = nm.array([[0, -.2], [.1, .2], [.3, .6], [.4, 1]])
    sp_l = BSpline(3, is_cyclic=False)
    sp_l.approximate(line_l, ncp=4)
    kn_lr = sp_l.get_knot_vector()
    sp_r = BSpline(3, is_cyclic=False)
    sp_r.approximate(line_r, knots=kn_lr)

    line_b = nm.array([[-1, -.5], [-.8, -.6], [-.5, -.4], [-.2, -.2],
                       [0, -.2]])
    line_t = nm.array([[.4, 1], [0, 1], [-.2, 1], [-.6, 1], [-1, 1]])
    sp_b = BSpline(3, is_cyclic=False)
    sp_b.approximate(line_b, ncp=5)
    kn_bt = sp_b.get_knot_vector()
    sp_t = BSpline(3, is_cyclic=False)
    sp_t.approximate(line_t, knots=kn_bt)

    mesh = Mesh.from_file(data_dir + '/meshes/2d/square_tri1.mesh')
    spb = SplineRegion2D([sp_b, sp_r, sp_t, sp_l], mesh.coors)
    spb.move_control_point(5, [-.2, .1])
    spb.move_control_point(10, [-.3, .2])
    spb.move_control_point(15, [-.1, .2])

    pt0 = mesh.coors[145,:].copy()
    mesh.cmesh.coors[:] = spb.evaluate()
    pt1 = mesh.coors[145,:]

    expected_distance = 0.0908306614584
    actual_distance = nm.linalg.norm(pt0 - pt1)
    ok = nm.fabs(actual_distance - expected_distance)\
        / expected_distance < tolerance

    if not ok:
        tst.report('expected distance:')
        tst.report(expected_distance)
        tst.report('actual distance:')
        tst.report(actual_distance)

    assert ok



import os.path as op

import numpy as nm
import pytest

import sfepy.base.testing as tst
from sfepy import data_dir

def init_vec(variables):
    return nm.random.rand(variables.di.n_dof_total)

def check_vec(vec, ii, ok, conds, variables):
    from sfepy.discrete.common.dof_info import expand_nodes_to_equations

    for var_name, var_conds in conds.group_by_variables().items():
        var = variables[var_name]
        for cond in var_conds:
            cond.canonize_dof_names(var.dofs)
            tst.report('%d: %s %s: %s %s'
                       % (ii, var.name,
                          cond.name, cond.region.name, cond.dofs[0]))
            nods = var.field.get_dofs_in_region(cond.region)
            eq = expand_nodes_to_equations(nods, cond.dofs[0], var.dofs)

            off = variables.di.indx[var_name].start
            n_nod = len(nods)
            for cdof, dof_name in enumerate(cond.dofs[0]):
                idof = var.dofs.index(dof_name)
                eqs = eq[n_nod * cdof : n_nod * (cdof + 1)]

                _ok = nm.allclose(vec[off + eqs], idof,
                                  atol=1e-14, rtol=0.0)
                if not _ok:
                    tst.report(' %s: failed! (all of %s == %f)'
                               % (dof_name, vec[off + eqs], idof))
                ok = ok and _ok

    return ok

@pytest.fixture(scope='module')
def data():
    from sfepy.base.base import Struct
    from sfepy.discrete import FieldVariable, Variables, Problem
    from sfepy.discrete.fem import Mesh, FEDomain, Field

    mesh = Mesh.from_file(data_dir + '/meshes/2d/square_unit_tri.mesh')
    domain = FEDomain('domain', mesh)

    omega = domain.create_region('Omega', 'all')
    domain.create_region('Left',
                         'vertices in (x < -0.499)',
                         'facet')
    domain.create_region('LeftStrip',
                         'vertices in (x < -0.499)'
                         ' & (y > -0.199) & (y < 0.199)',
                         'facet')
    domain.create_region('LeftFix',
                         'r.Left -v r.LeftStrip',
                         'facet')
    domain.create_region('Right',
                         'vertices in (x > 0.499)',
                         'facet')
    domain.create_region('RightStrip',
                         'vertices in (x > 0.499)'
                         ' & (y > -0.199) & (y < 0.199)',
                         'facet')
    domain.create_region('RightFix',
                         'r.Right -v r.RightStrip',
                         'facet')

    fu = Field.from_args('fu', nm.float64, 'vector', omega, approx_order=2)
    u = FieldVariable('u', 'unknown', fu)

    fp = Field.from_args('fp', nm.float64, 'scalar', omega, approx_order=2)
    p = FieldVariable('p', 'unknown', fp)

    pb = Problem('test', domain=domain, fields=[fu, fp],
                 auto_conf=False)

    return Struct(problem=pb, variables=Variables([u, p]))

def test_ics(data):
    from sfepy.discrete.conditions import Conditions, InitialCondition

    variables = data.variables
    omega = data.problem.domain.regions['Omega']

    all_ics = []
    all_ics.append(InitialCondition('ic0', omega,
                                    {'p.all' : 0.0}))
    all_ics.append(InitialCondition('ic1', omega,
                                    {'u.1' : 1.0}))
    all_ics.append(InitialCondition('ic2', omega,
                                    {'u.all' : nm.array([0.0, 1.0])}))
    all_ics.append(InitialCondition('ic3', omega,
                                    {'p.0' : 0.0,
                                     'u.0' : 0.0, 'u.1' : 1.0}))

    ok = True
    for ii, ics in enumerate(all_ics):
        if not isinstance(ics, list): ics = [ics]

        ics = Conditions(ics)
        variables.setup_initial_conditions(ics, functions=None)

        vec = init_vec(variables)
        variables.apply_ic(vec)

        ok = check_vec(vec, ii, ok, ics, variables)

    assert ok

def test_ebcs(data):
    from sfepy.discrete.conditions import Conditions, EssentialBC

    variables = data.variables
    regions = data.problem.domain.regions

    all_ebcs = []
    all_ebcs.append(EssentialBC('fix_u1', regions['LeftFix'],
                                {'u.all' : nm.array([0.0, 1.0])}))
    all_ebcs.append(EssentialBC('fix_u2', regions['LeftStrip'],
                                {'u.0' : 0.0, 'u.1' : 1.0}))
    all_ebcs.append(EssentialBC('fix_p1', regions['RightFix'],
                                {'p.all' : 0.0}))
    all_ebcs.append(EssentialBC('fix_p2', regions['RightStrip'],
                                {'p.0' : 0.0}))
    all_ebcs.append([EssentialBC('fix_p3', regions['Right'],
                                 {'p.0' : 0.0}),
                     EssentialBC('fix_u3', regions['Left'],
                                 {'u.0' : 0.0, 'u.1' : 1.0})])

    ok = True
    for ii, bcs in enumerate(all_ebcs):
        if not isinstance(bcs, list): bcs = [bcs]

        ebcs = Conditions(bcs)
        variables.equation_mapping(ebcs=ebcs, epbcs=None,
                                   ts=None, functions=None)
        vec = init_vec(variables)
        variables.apply_ebc(vec)

        ok = check_vec(vec, ii, ok, ebcs, variables)

    assert ok

def test_epbcs(data):
    from sfepy.discrete import Function, Functions
    from sfepy.discrete.conditions import Conditions, PeriodicBC
    from sfepy.discrete.common.dof_info import expand_nodes_to_equations
    from sfepy.discrete.fem.periodic import match_y_line

    variables = data.variables
    regions = data.problem.domain.regions

    match_y_line = Function('match_y_line', match_y_line)
    pbc = PeriodicBC('pbc', [regions['LeftStrip'], regions['RightStrip']],
                     {'u.[1,0]' : 'u.[0,1]'}, match='match_y_line')

    functions = Functions([match_y_line])

    epbcs = Conditions([pbc])
    variables.equation_mapping(ebcs=None, epbcs=epbcs,
                               ts=None, functions=functions)

    vec = init_vec(variables)
    variables.apply_ebc(vec)

    var = variables['u']
    var_bcs = epbcs.group_by_variables()['u']
    bc = var_bcs['pbc']
    bc.canonize_dof_names(var.dofs)

    nods0 = var.field.get_dofs_in_region(bc.regions[0])
    nods1 = var.field.get_dofs_in_region(bc.regions[1])

    coors0 = var.field.get_coor(nods0)
    coors1 = var.field.get_coor(nods1)
    i0, i1 = match_y_line(coors0, coors1)

    eq0 = expand_nodes_to_equations(nods0[i0], bc.dofs[0], var.dofs)
    eq1 = expand_nodes_to_equations(nods1[i1], bc.dofs[1], var.dofs)

    ok = True

    _ok = variables.has_ebc(vec)
    if not _ok:
        tst.report('EPBCs were not applied correctly!')
    ok = ok and _ok

    _ok = len(nm.setdiff1d(eq0, var.eq_map.master)) == 0
    if not _ok:
        tst.report('master equations mismatch! (set(%s) == set(%s))'
                   % (eq0, var.eq_map.master))
    ok = ok and _ok

    _ok = len(nm.setdiff1d(eq1, var.eq_map.slave)) == 0
    if not _ok:
        tst.report('slave equations mismatch! (set(%s) == set(%s))'
                   % (eq1, var.eq_map.slave))
    ok = ok and _ok

    off = variables.di.indx['u'].start
    _ok = nm.allclose(vec[off + eq0], vec[off + eq1], atol=1e-14, rtol=0.0)
    if not _ok:
        tst.report('periodicity test failed! (%s == %s)'
                   % (vec[off + eq0], vec[off + eq1]))
    ok = ok and _ok

    assert ok

def test_save_ebc(data, output_dir):
    from sfepy.discrete import (FieldVariable, Integral,
                                Equation, Equations, Problem)
    from sfepy.discrete.conditions import Conditions, EssentialBC
    from sfepy.terms import Term

    name = op.join(output_dir,
                   op.splitext(op.basename(__file__))[0])

    integral = Integral('i', order=1)

    u = data.variables['u']
    v = FieldVariable('v', 'test', u.field, primary_var_name='u')

    p = data.variables['p']
    q = FieldVariable('q', 'test', p.field, primary_var_name='p')

    regions = data.problem.domain.regions
    omega = regions['Omega']

    # Problem.save_ebc() requires to have equations defined.
    t1 = Term.new('dw_lin_elastic(v, u)',
                  integral, omega, v=v, u=u)
    t2 = Term.new('dw_laplace(q, p)', integral, omega, q=q, p=p)
    eq = Equation('aux', t1 + t2)
    eqs = Equations([eq])

    pb = Problem('test', equations=eqs)

    all_ebcs = []
    all_ebcs.append(EssentialBC('fix_u1', regions['RightFix'],
                                {'u.all' : nm.array([0.0, 1.0])}))
    all_ebcs.append(EssentialBC('fix_u2', regions['LeftStrip'],
                                {'u.0' : 0.0, 'u.1' : 1.0}))
    all_ebcs.append(EssentialBC('fix_p1', regions['LeftFix'],
                                {'p.all' : 0.0}))
    all_ebcs.append(EssentialBC('fix_p2', regions['RightStrip'],
                                {'p.0' : 0.0}))
    ebcs = Conditions(all_ebcs)

    pb.time_update(ebcs=ebcs)

    pb.save_ebc(name + '_ebcs_f.vtk', ebcs=ebcs, force=True)
    pb.save_ebc(name + '_ebcs.vtk', ebcs=ebcs, default=-1, force=False)

    assert True



import os
import pytest
from sfepy.base.testing import check_conditions, run_declaratice_example

# Relative to sfepy.base_dir, see run_declaratice_example().
examples_dir = 'examples/'

def inedir(filename):
    return os.path.join(examples_dir, filename)

examples = [
    'acoustics/acoustics.py',
    'acoustics/acoustics3d.py',
    'acoustics/helmholtz_apartment.py',
    'acoustics/vibro_acoustic3d.py',
    'diffusion/cube.py',
    'diffusion/darcy_flow_multicomp.py',
    'diffusion/laplace_1d.py',
    'diffusion/laplace_coupling_lcbcs.py',
    'diffusion/laplace_fluid_2d.py',
    'diffusion/laplace_time_ebcs.py',
    'diffusion/poisson.py',
    'diffusion/poisson_field_dependent_material.py',
    'diffusion/poisson_functions.py',
    'diffusion/poisson_neumann.py',
    'diffusion/poisson_nonlinear_material.py',
    'diffusion/poisson_periodic_boundary_condition.py',
    'diffusion/sinbc.py',
    'diffusion/time_advection_diffusion.py',
    'diffusion/time_heat_equation_multi_material.py',
    'diffusion/time_poisson.py',
    'homogenization/linear_elastic_mM.py',
    'large_deformation/active_fibres.py',
    'large_deformation/balloon.py',
    'large_deformation/perfusion_tl.py',
    'linear_elasticity/elastic_contact_planes.py',
    'linear_elasticity/elastic_contact_sphere.py',
    'linear_elasticity/elastic_shifted_periodic.py',
    'linear_elasticity/elastodynamic.py',
    'linear_elasticity/its2D_2.py',
    'linear_elasticity/linear_elastic.py',
    'linear_elasticity/linear_elastic_damping.py',
    'linear_elasticity/linear_elastic_probes.py',
    'linear_elasticity/linear_elastic_tractions.py',
    'linear_elasticity/linear_elastic_up.py',
    'linear_elasticity/linear_viscoelastic.py',
    'linear_elasticity/material_nonlinearity.py',
    'linear_elasticity/mixed_mesh.py',
    'linear_elasticity/modal_analysis_declarative.py',
    'linear_elasticity/multi_node_lcbcs.py',
    'linear_elasticity/nodal_lcbcs.py',
    'linear_elasticity/prestress_fibres.py',
    'linear_elasticity/seismic_load.py',
    'linear_elasticity/shell10x_cantilever.py',
    'linear_elasticity/truss_bridge3d.py',
    'linear_elasticity/truss_bridge.py',
    'linear_elasticity/two_bodies_contact.py',
    'linear_elasticity/wedge_mesh.py',
    'multi_physics/biot.py',
    'multi_physics/biot_npbc.py',
    'multi_physics/biot_npbc_lagrange.py',
    'multi_physics/biot_short_syntax.py',
    'multi_physics/piezo_elasticity.py',
    'multi_physics/piezo_elastodynamic.py',
    'multi_physics/thermo_elasticity.py',
    'multi_physics/thermo_elasticity_ess.py',
    'navier_stokes/navier_stokes.py',
    'navier_stokes/navier_stokes2d.py',
    'navier_stokes/stabilized_navier_stokes.py',
    'navier_stokes/stokes.py',
    'navier_stokes/stokes_slip_bc.py',
    'quantum/boron.py',
    'quantum/hydrogen.py',
    'quantum/oscillator.py',
    'quantum/well.py',
]

try:
    from igakit import igalib; igalib

except ImportError:
    pass

else:
    examples.extend([
        'diffusion/poisson_iga.py',
        'linear_elasticity/linear_elastic_iga.py',
        'navier_stokes/navier_stokes2d_iga.py',
    ])

@pytest.mark.parametrize('ex_filename', examples)
def test_examples(ex_filename, output_dir):
    conditions = run_declaratice_example(
        ex_filename=inedir(ex_filename),
        output_dir=output_dir,
        remove_prefix=examples_dir,
    )
    if ex_filename == 'large_deformation/active_fibres.py':
        # Special-case the first iteration, as the solver converges slowly.
        ok = (conditions[1:] == 0).all()
        ok = ok and (conditions[0] == 1)

    elif ex_filename == 'large_deformation/balloon.py':
        # Special-case the steps with a time-step reduction.
        ok = (conditions[1] == 1)
        ok = ok and (conditions[0] == 0) and (conditions[2:] == 0).all()

    else:
        ok = check_conditions(conditions)

    assert ok

examples_dg = [
    'dg/advection_1D.py',
    'dg/advection_2D.py',
    'dg/advection_diffusion_2D.py',
    'dg/laplace_2D.py',
]
@pytest.mark.parametrize('ex_filename', examples_dg)
def test_examples_dg(ex_filename, output_dir):
    conditions = run_declaratice_example(
        ex_filename=inedir(ex_filename),
        output_dir=output_dir,
        ext='.msh',
        remove_prefix=examples_dir,
    )
    ok = check_conditions(conditions)
    assert ok



import sfepy.base.testing as tst

def test_compose_sparse():
    import numpy as nm
    import scipy.sparse as sps
    from sfepy.linalg import compose_sparse

    ok = True

    # basic
    ma = sps.csr_matrix([[1, 0], [0, 1]])
    mb = sps.coo_matrix([[1, 1]])
    mk = compose_sparse([[ma, mb.T], [mb, 0]])
    expected = nm.array([[1, 0, 1],
                         [0, 1, 1],
                         [1, 1, 0]])

    _ok = nm.all(mk.toarray() == expected)
    tst.report('basic: %s' % _ok)
    ok = ok and _ok

    # sizes and slices
    ma = sps.csr_matrix([[2, 3]])
    mb = sps.coo_matrix([[4, 5, 6]])

    mk = compose_sparse([[ma, mb]], col_sizes=[2, 3])
    expected = nm.array([[2, 3, 4, 5, 6]])

    _ok = nm.all(mk.toarray() == expected)
    tst.report('sizes: %s' % _ok)
    ok = ok and _ok

    i1 = slice(1, 3)
    i2 = slice(8, 11)
    mk = compose_sparse([[ma, mb]], col_sizes=[i1, i2])
    expected = nm.array([[0, 2, 3, 0, 0, 0, 0, 0, 4, 5, 6]])

    _ok = nm.all(mk.toarray() == expected)
    tst.report('slices: %s' % _ok)
    ok = ok and _ok

    # zero block sizes and slices
    mk = compose_sparse([[0, ma, 0, mb, 0]], col_sizes=[1, 2, 5, 3, 1])
    expected = nm.array([[0, 2, 3, 0, 0, 0, 0, 0, 4, 5, 6, 0]])

    _ok = nm.all(mk.toarray() == expected)
    tst.report('zero block sizes: %s' % _ok)
    ok = ok and _ok

    expected = nm.array([[0, 2, 3, 0, 4, 5, 6, 0, 0, 0, 0],
                         [0, 0, 0, 0, 0, 0, 0, 0, 4, 5, 6]])

    i0 = slice(0, 1)
    i1 = slice(1, 3)
    i2 = slice(4, 7)
    i3 = slice(8, 11)
    mk = compose_sparse([[0, ma, mb, 0],
                         [0, 0, 0, mb]], col_sizes=[i0, i1, i2, i3])

    _ok = nm.all(mk.toarray() == expected)
    tst.report('zero block slices: %s' % _ok)
    ok = ok and _ok

    assert ok



r"""
Vibro-acoustic problem

3D acoustic domain with 2D perforated deforming interface.

Problem definition - find :math:`p` (acoustic pressure),
:math:`g` (transversal acoustic velocity),
:math:`w` (plate deflection) and :math:`\ul{\theta}` (rotation) such that:

.. math::
    c^2 \int_{\Omega} \nabla q \cdot \nabla p
    - \omega^2 \int_{\Omega} q p
    + i \omega c \int_{\Gamma_{in}} q p
    + i \omega c \int_{\Gamma_{out}} q p
    - i \omega c^2 \int_{\Gamma_0} (q^+ - q^-) g
    = 2i \omega c \int_{\Gamma_{in}} q \bar{p}
    \;, \quad \forall q \;,

    - i \omega \int_{\Gamma_0} f (p^+ - p^-)
    - \omega^2 \int_{\Gamma_0} F f g
    + \omega^2 \int_{\Gamma_0} C f w
    = 0
    \;, \quad \forall f \;,

    \omega^2 \int_{\Gamma_0} C z g
    - \omega^2 \int_{\Gamma_0} S z w
    + \int_{\Gamma_0} \nabla z \cdot \ull{G} \cdot \nabla w
    - \int_{\Gamma_0} \ul{\theta} \cdot \ull{G} \cdot \nabla z
    = 0
    \;, \quad \forall z \;,

    - \omega^2 \int_{\Gamma_0} R\, \ul{\nu} \cdot \ul{\theta}
    + \int_{\Gamma_0} D_{ijkl} e_{ij}(\ul{\nu}) e_{kl}(\ul{\theta})
    - \int_{\Gamma_0} \ul{\nu} \cdot \ull{G} \cdot \nabla w
    + \int_{\Gamma_0} \ul{\nu} \cdot \ull{G} \cdot \ul{\theta}
    = 0
    \;, \quad \forall \ul{\nu} \;,
"""
import numpy as nm
from sfepy import data_dir
from sfepy.mechanics.matcoefs import stiffness_from_lame


def define(sound_speed=343.0,
           wave_num=5.5,
           p_inc=300,
           thickness=0.01,
           filename_mesh='/meshes/3d/acoustic_wg.vtk'):

    filename_mesh = data_dir + filename_mesh

    c = sound_speed
    c2 = c**2
    w = wave_num * c
    w2 = w**2
    wc = w * c
    wc2 = w * c2

    regions = {
        'Omega1': 'cells of group 1',
        'Omega2': 'cells of group 2',
        'GammaIn': ('vertices of group 1', 'face'),
        'GammaOut': ('vertices of group 2', 'face'),
        'Gamma_aux': ('r.Omega1 *v r.Omega2', 'face'),
        'Gamma0_1': ('copy r.Gamma_aux', 'face', 'Omega1'),
        'Gamma0_2': ('copy r.Gamma_aux', 'face', 'Omega2'),
        'Gamma0': ('copy r.Gamma_aux', 'cell', None, {'mesh_dim': 2}),
        'Left_': ('vertices in (x < 0.001)', 'edge'),
        'Right_': ('vertices in (x > 0.299)', 'edge'),
        'Gamma0_Left': ('r.Gamma_aux *v r.Left_', 'edge'),
        'Gamma0_Right': ('r.Gamma_aux *v r.Right_', 'edge'),
        }

    fields = {
        'pressure1': ('complex', 'scalar', 'Omega1', 1),
        'pressure2': ('complex', 'scalar', 'Omega2', 1),
        'tvelocity': ('complex', 'scalar', 'Gamma0', 1),
        'deflection': ('complex', 'scalar', 'Gamma0', 1),
        'rotation': ('complex', 'vector', 'Gamma0', 1),
        }

    variables = {
        'p1': ('unknown field', 'pressure1', 0),
        'q1': ('test field', 'pressure1', 'p1'),
        'p2': ('unknown field', 'pressure2', 1),
        'q2': ('test field', 'pressure2', 'p2'),
        'g0': ('unknown field', 'tvelocity', 2),
        'f0': ('test field', 'tvelocity', 'g0'),
        'w': ('unknown field', 'deflection', 3),
        'z': ('test field', 'deflection', 'w'),
        'theta': ('unknown field', 'rotation', 4),
        'nu': ('test field', 'rotation', 'theta'),
        }

    ebcs = {
        'fixed_l': ('Gamma0_Left', {'w.0': 0.0, 'theta.all': 0.0}),
        'fixed_r': ('Gamma0_Right', {'w.0': 0.0, 'theta.all': 0.0}),
        }

    options = {
        'split_results_by': 'region',
        'output_dir': 'output',
        }

    functions = {
        }

    materials = {
        'ac': ({
            'F': -2.064e+00,
            'c': -1.064e+00,
            'T': 9.202e-01,
            'hG': thickness * 4.5e10 * nm.eye(2),
            'hR': thickness * 0.71,
            'h3R': thickness**3 / 3.0 * 0.71,
            'h3C': thickness**3 / 3.0 * stiffness_from_lame(2, 1e1, 1e0)}, ),
        }

    equations = {
        'eq_p1': """
            %e * dw_laplace.5.Omega1(q1, p1)
        - %e * dw_dot.5.Omega1(q1, p1)
        + %s * dw_dot.5.GammaIn(q1, p1)
        - %s * dw_dot.5.Gamma0_1(q1, tr(Gamma0, g0))
        = %s * dw_integrate.5.GammaIn(q1)""" % (c2, w2, 1j * wc,
                                                1j * wc2, 2j * wc * p_inc),
        'eq_p2': """
        + %e * dw_laplace.5.Omega2(q2, p2)
        - %e * dw_dot.5.Omega2(q2, p2)
        + %s * dw_dot.5.GammaOut(q2, p2)
        + %s * dw_dot.5.Gamma0_2(q2, tr(Gamma0, g0))
        = 0""" % (c2, w2, 1j * wc, 1j * wc2),
        'eq_g0': """
        - %s * dw_dot.5.Gamma0(f0, tr(Gamma0_1, p1))
        + %s * dw_dot.5.Gamma0(f0, tr(Gamma0_2, p2))
        - %e * dw_dot.5.Gamma0(ac.F, f0, g0)
        + %e * dw_dot.5.Gamma0(ac.c, f0, w)
        = 0""" % (1j * w, 1j * w, w2, w2),
        'eq_w': """
            %e * dw_dot.5.Gamma0(ac.c, z, g0)
        - %e * dw_dot.5.Gamma0(ac.T, z, w)
        - %e * dw_dot.5.Gamma0(ac.hR, z, w)
            + dw_diffusion.5.Gamma0(ac.hG, z, w)
            - dw_v_dot_grad_s.5.Gamma0(ac.hG, theta, z)
        = 0""" % (w2, w2, w2),
        'eq_theta': """
        - %e * dw_dot.5.Gamma0(ac.h3R, nu, theta)
            + dw_lin_elastic.5.Gamma0(ac.h3C, nu, theta)
            - dw_v_dot_grad_s.5.Gamma0(ac.hG, nu, w)
            + dw_dot.5.Gamma0(ac.hG, nu, theta)
        = 0""" % (w2, ),
        }

    solvers = {
        'ls': ('ls.auto_direct', {}),
        'nls': ('nls.newton', {
            'i_max': 1,
            'eps_a': 1e-4,
            'eps_r': 1e-6,
        })
    }

    return locals()



"""
Time stepping solvers.
"""
from inspect import signature
from functools import partial
import os.path as osp
import numpy as nm

from sfepy.base.base import (get_default, output, assert_, Struct)
from sfepy.base.timing import Timer
from sfepy.linalg.utils import output_array_stats
from sfepy.solvers.solvers import TimeSteppingSolver, NonlinearSolver
from sfepy.solvers.ls import RMMSolver
from sfepy.solvers.ts_controllers import FixedTSC
from sfepy.solvers.ts import TimeStepper, VariableTimeStepper

def standard_ts_call(call):
    """
    Decorator handling argument preparation and timing for time-stepping
    solvers.
    """

    def _standard_ts_call(self, vec0=None, nls=None,
                         init_fun=None, prestep_fun=None, poststep_fun=None,
                         status=None, log_nls_status=False, **kwargs):
        timer = Timer(start=True)

        nls = get_default(nls, self.nls,
                          'nonlinear solver has to be specified!')

        init_fun = get_default(init_fun, lambda ts, vec0: vec0)
        prestep_fun = get_default(prestep_fun, lambda ts, vec: None)
        poststep_fun = get_default(poststep_fun, lambda ts, vec: None)

        _nls = nls
        if status is not None:
            nls_status = status.get('nls_status')
            if nls_status is not None:
                if log_nls_status:
                    pb = self.context
                    filename_log = osp.join(pb.output_dir,
                                            pb.ofn_trunk + '_log.csv')
                    status.log_file = open(filename_log, 'wt',
                                           encoding="utf-8")

                class _TimingNLS(type(nls)):
                    def __call__(self, *args, **kwargs):
                        # Call the original nls...
                        out = super().__call__(*args, **kwargs)

                        log_stats = {}
                        # ...and collect its time stats.
                        time_stats = nls_status.get('time_stats')
                        if time_stats is not None:
                            all_stats = status.setdefault('time_stats', {})
                            for key, val in time_stats.items():
                                all_stats.setdefault(key, 0.0)
                                all_stats[key] += val

                            log_stats.update(time_stats)

                            all_stats.setdefault('time', 0.0)
                            all_stats['time'] += nls_status.get('time', 0.0)

                        # ...and collect its step stats.
                        all_stats = status.setdefault('step_stats', [])
                        _nls_status = nls_status.copy()
                        _nls_status.step = self.context.ts.step
                        _nls_status.step_time = self.context.ts.time
                        all_stats.append(_nls_status)
                        log_stats.update(_nls_status.to_dict())

                        if getattr(status, 'log_file', None) is not None:
                            if 'time_stats' in log_stats:
                                del log_stats['time_stats']

                            keys = list(log_stats.keys())
                            keys.sort()

                            if status.log_file.tell() == 0:
                                status.log_file.write(','.join(keys) + '\n')

                            log_vals = [f'{log_stats[key]}' for key in keys]
                            status.log_file.write(','.join(log_vals) + '\n')
                            status.log_file.flush()

                        return out

                _nls = nls.copy()
                _nls.__class__ = _TimingNLS

        result = call(self, vec0=vec0, nls=_nls, init_fun=init_fun,
                      prestep_fun=prestep_fun, poststep_fun=poststep_fun,
                      status=status, **kwargs)

        elapsed = timer.stop()
        if status is not None:
            status['time'] = elapsed
            status['n_step'] = self.ts.n_step
            if getattr(status, 'log_file', None) is not None:
                status.log_file.close()

        return result

    return _standard_ts_call

#
# General solvers.
#
class StationarySolver(TimeSteppingSolver):
    """
    Solver for stationary problems without time stepping.

    This class is provided to have a unified interface of the time stepping
    solvers also for stationary problems.
    """
    name = 'ts.stationary'

    def __init__(self, conf, nls=None, context=None, **kwargs):
        TimeSteppingSolver.__init__(self, conf, nls=nls, context=context,
                                    **kwargs)

        self.ts = TimeStepper(0.0, 1.0, n_step=1, is_quasistatic=True)

    @standard_ts_call
    def __call__(self, vec0=None, nls=None, init_fun=None, prestep_fun=None,
                 poststep_fun=None, status=None, **kwargs):
        ts = self.ts
        nls = get_default(nls, self.nls)

        vec0 = init_fun(ts, vec0)

        vec0 = prestep_fun(ts, vec0)

        vec = nls(vec0)

        vec = poststep_fun(ts, vec)

        return vec

class SimpleTimeSteppingSolver(TimeSteppingSolver):
    """
    Implicit time stepping solver with a fixed time step.
    """
    name = 'ts.simple'

    _parameters = [
        ('t0', 'float', 0.0, False,
         'The initial time.'),
        ('t1', 'float', 1.0, False,
         'The final time.'),
        ('dt', 'float', None, False,
         'The time step. Used if `n_step` is not given.'),
        ('n_step', 'int', 10, False,
         'The number of time steps. Has precedence over `dt`.'),
        ('quasistatic', 'bool', False, False,
         """If True, assume a quasistatic time-stepping. Then the non-linear
            solver is invoked also for the initial time."""),
    ]

    def __init__(self, conf, nls=None, context=None, **kwargs):
        TimeSteppingSolver.__init__(self, conf, nls=nls, context=context,
                                    **kwargs)
        self.ts = TimeStepper.from_conf(self.conf)

        nd = self.ts.n_digit
        format = '====== time %%e (step %%%dd of %%%dd) =====' % (nd, nd)

        self.format = format
        self.verbose = self.conf.verbose

    def solve_step0(self, nls, vec0):
        if self.conf.quasistatic:
            vec = nls(vec0)

        else:
            res = nls.fun(vec0)
            err = nm.linalg.norm(res)
            output('initial residual: %e' % err, verbose=self.verbose)
            vec = vec0.copy()

        return vec

    def solve_step(self, ts, nls, vec, prestep_fun=None):
        return nls(vec)

    def output_step_info(self, ts):
        output(self.format % (ts.time, ts.step + 1, ts.n_step),
               verbose=self.verbose)

    @standard_ts_call
    def __call__(self, vec0=None, nls=None, init_fun=None, prestep_fun=None,
                 poststep_fun=None, status=None, **kwargs):
        """
        Solve the time-dependent problem.
        """
        ts = self.ts
        nls = get_default(nls, self.nls)

        vec0 = init_fun(ts, vec0)

        self.output_step_info(ts)
        if ts.step == 0:
            vec0 = prestep_fun(ts, vec0)

            vec = self.solve_step0(nls, vec0)

            vec = poststep_fun(ts, vec)
            ts.advance()

        else:
            vec = vec0

        for step, time in ts.iter_from(ts.step):
            self.output_step_info(ts)

            vec = prestep_fun(ts, vec)

            vect = self.solve_step(ts, nls, vec, prestep_fun)

            vect = poststep_fun(ts, vect)

            vec = vect

        return vec

def get_min_dt(adt):
    red = adt.red
    while red >= adt.red_max:
        red *= adt.red_factor

    dt = adt.dt0 * red

    return dt

def adapt_time_step(ts, status, adt, context=None, verbose=False):
    """
    Adapt the time step of `ts` according to the exit status of the
    nonlinear solver.

    The time step dt is reduced, if the nonlinear solver did not converge. If it
    converged in less then a specified number of iterations for several time
    steps, the time step is increased. This is governed by the following
    parameters:

    - `red_factor` : time step reduction factor
    - `red_max` : maximum time step reduction factor
    - `inc_factor` : time step increase factor
    - `inc_on_iter` : increase time step if the nonlinear solver converged in
      less than this amount of iterations...
    - `inc_wait` : ...for this number of consecutive time steps

    Parameters
    ----------
    ts : VariableTimeStepper instance
        The time stepper.
    status : IndexedStruct instance
        The nonlinear solver exit status.
    adt : Struct instance
        The object with the adaptivity parameters of the time-stepping solver
        such as `red_factor` (see above) as attributes.
    context : object, optional
        The context can be used in user-defined adaptivity functions. Not used
        here.

    Returns
    -------
    is_break : bool
        If True, the adaptivity loop should stop.
    """
    is_break = False

    if status.condition == 0:
        if status.n_iter <= adt.inc_on_iter:
            adt.wait += 1

            if adt.wait > adt.inc_wait:
                if adt.red < 1.0:
                    adt.red = adt.red * adt.inc_factor
                    ts.set_time_step(adt.dt0 * adt.red)
                    output('+++++ new time step: %e +++++' % ts.dt,
                           verbose=verbose)
                adt.wait = 0

        else:
            adt.wait = 0

        is_break = True

    else:
        adt.red = adt.red * adt.red_factor
        if adt.red < adt.red_max:
            is_break = True

        else:
            ts.set_time_step(adt.dt0 * adt.red, update_time=True)
            output('----- new time step: %e -----' % ts.dt,
                   verbose=verbose)
            adt.wait = 0

    return is_break

class AdaptiveTimeSteppingSolver(SimpleTimeSteppingSolver):
    """
    Implicit time stepping solver with an adaptive time step.

    Either the built-in or user supplied function can be used to adapt the time
    step.
    """
    name = 'ts.adaptive'

    _parameters = SimpleTimeSteppingSolver._parameters + [
        ('adapt_fun', 'callable(ts, status, adt, context, verbose)',
         None, False,
         """If given, use this function to set the time step in `ts`. The
            function return value is a bool - if True, the adaptivity loop
            should stop. The other parameters below are collected in `adt`,
            `status` is the nonlinear solver status, `context` is
            a user-defined context and `verbose` is a verbosity flag.
            Solvers created by
            :class:`Problem <sfepy.discrete.problem.Problem>` use the
            Problem instance as the context."""),
        ('dt_red_factor', 'float', 0.2, False,
         'The time step reduction factor.'),
        ('dt_red_max', 'float', 1e-3, False,
         'The maximum time step reduction factor.'),
        ('dt_inc_factor', 'float', 1.25, False,
         'The time step increase factor.'),
        ('dt_inc_on_iter', 'int', 4, False,
         """Increase the time step if the nonlinear solver converged in less
            than this amount of iterations for `dt_inc_wait` consecutive time
            steps."""),
        ('dt_inc_wait', 'int', 5, False,
         'The number of consecutive time steps, see `dt_inc_on_iter`.'),
    ]

    def __init__(self, conf, nls=None, context=None, **kwargs):
        TimeSteppingSolver.__init__(self, conf, nls=nls, context=context,
                                    **kwargs)

        self.ts = VariableTimeStepper.from_conf(self.conf)

        get = self.conf.get
        adt = Struct(red_factor=get('dt_red_factor', 0.2),
                     red_max=get('dt_red_max', 1e-3),
                     inc_factor=get('dt_inc_factor', 1.25),
                     inc_on_iter=get('dt_inc_on_iter', 4),
                     inc_wait=get('dt_inc_wait', 5),
                     red=1.0, wait=0, dt0=0.0)
        self.adt = adt

        adt.dt0 = self.ts.get_default_time_step()
        self.ts.set_n_digit_from_min_dt(get_min_dt(adt))

        self.format = '====== time %e (dt %e, wait %d, step %d of %d) ====='
        self.verbose = self.conf.verbose

        self.adapt_time_step = self.conf.adapt_fun
        if self.adapt_time_step is None:
            self.adapt_time_step = adapt_time_step

    def solve_step(self, ts, nls, vec, prestep_fun):
        """
        Solve a single time step.
        """
        while 1:
            vect = nls(vec, status=nls.status)

            is_break = self.adapt_time_step(ts, nls.status, self.adt,
                                            self.context, verbose=self.verbose)

            if is_break:
                break

            vec = prestep_fun(ts, vec)

        return vect

    def output_step_info(self, ts):
        output(self.format % (ts.time, ts.dt, self.adt.wait,
                              ts.step + 1, ts.n_step),
               verbose=self.verbose)

#
# Elastodynamics solvers.
#
def transform_equations_ed(equations, materials):
    """
    Transform equations and variables for :class:`ElastodynamicsBaseTS`-based
    time stepping solvers. The displacement variable name is automatically
    detected by seeking the second time derivative, i.e. the 'dd' prefix in
    variable names.
    """
    from sfepy.terms.terms import Terms, Term
    from sfepy.discrete.variables import FieldVariable
    from sfepy.discrete.equations import Equations, Equation

    eq_mass = []
    eq_damping = []
    eq_other = []
    new_var_primary = set()
    variables = equations.variables
    for eq in equations:
        for term in eq.terms:
            for name in term.names.state:
                der = term.arg_derivatives[name]
                if ((der is not None) and
                    isinstance(der, int) and
                    (der > 0)):
                    new_var_primary.add(name)
                    if der == 2:
                        eq_mass.append(term)

                    else:
                        eq_damping.append(term)

                else:
                    eq_other.append(term)

                continue

    assert len(new_var_primary) == 1
    uname = new_var_primary.pop()
    vname = variables[uname].dual_var_name
    duname = 'd' + uname
    dduname = 'dd' + uname
    dvname = 'd' + vname
    ddvname = 'dd' + vname
    new_variables = variables.copy()
    if new_variables[uname]._order is None:
        raise ValueError('state variable orders have to be specified when using'
                         ' auto_transform_equations!')
    num = len(new_variables.state)
    new_variables.extend([
        FieldVariable(duname, 'unknown', variables[uname].field, order=num),
        FieldVariable(dduname, 'unknown', variables[uname].field, order=num + 1),
        FieldVariable(dvname, 'test', variables[uname].field,
                      primary_var_name=duname),
        FieldVariable(ddvname, 'test', variables[uname].field,
                      primary_var_name=dduname),
    ])
    for it, term in enumerate(eq_mass.copy()):
        aux = ','.join([ii.strip() for ii in term.arg_str.split(',')])
        arg_str = aux.replace(f',{vname},', f',{ddvname},')
        new_term = term.__class__(term.name, arg_str, term.integral, term.region)
        new_term.setup(allow_derivatives=False)
        new_term.assign_args(new_variables, materials, user=None)


        eq_mass[it] = new_term

    for it, term in enumerate(eq_damping.copy()):
        aux = ','.join([ii.strip() for ii in term.arg_str.split(',')])
        arg_str = aux.replace(f',{vname},', f',{dvname},')
        new_term = term.__class__(term.name, arg_str, term.integral, term.region)
        new_term.setup(allow_derivatives=False)
        new_term.assign_args(new_variables, materials, user=None)


        eq_damping[it] = new_term

    if not len(eq_damping):
        mterm = eq_mass[0]
        # Dummy damping to introduce du, could use a single cell region?
        dterm = Term.new(
            f'dw_zero({dvname}, {duname})', mterm.integral, mterm.region,
            **{dvname : new_variables[dvname],
               duname : new_variables[duname]},
        )
        dterm.setup(allow_derivatives=False)
        dterm.assign_args(new_variables, materials, user=None)
        eq_damping = [dterm]

    new_equations = Equations([Equation('M', Terms(eq_mass), setup=False),
                               Equation('C', Terms(eq_damping), setup=False),
                               Equation('K', Terms(eq_other), setup=False),])

    var_names = {
        'u' : uname, 'du' : duname, 'ddu' : dduname,
        'extra' : set(equations.variables.di.var_names)
        - set([uname, duname, dduname])
    }
    return new_equations, var_names

def gen_multi_vec_packing(di, names, extra_variables=False):
    """
    Return DOF vector (un)packing functions for nlst. For multiphysical
    problems (non-empty `ie` slice for extra variables) the `unpack()` function
    accepts an additional argument `mode` that can be set to 'full' or 'nls'.

    The following DOF ordering must be obeyed:

    - The full DOF vector:

      | ``---iue---|-iv-|-ia-``
      | ``-iu-|-ie-|``
    """
    iu = di.indx[names['u']]
    iv = di.indx[names['du']]
    ia = di.indx[names['ddu']]
    ie = slice(iu.stop, iv.start)
    iue = slice(0, iv.start)
    assert_(iu.stop == di.n_dof[names['u']])
    assert_(iv.start == ie.stop)
    assert_(ia.start == iv.stop)

    if not extra_variables:
        assert_(ie.stop == ie.start)
        n_arg = 3

        def unpack(vec, mode=None):
            return vec[iu], vec[iv], vec[ia]

    else:
        n_arg = 4

        def unpack(vec, mode='full'):
            if mode == 'nls':
                return vec[iu], vec[ie]

            else:
                return vec[iu], vec[ie], vec[iv], vec[ia]

    def pack(*args):
        return nm.concatenate(args)

    indices = dict(indices=(iue, iu, ie, iv, ia),
                   n_dof=di.n_dof_total, n_uedof=ie.stop, n_arg=n_arg)
    unpack.__dict__.update(indices)
    pack.__dict__.update(indices)
    return unpack, pack

def _cache(obj, attr, dep):
    def decorate(fun):
        def new_fun(*args, **kwargs):
            if dep:
                val = getattr(obj, attr)
                if val is None:
                    val = fun(*args, **kwargs)
                    setattr(obj, attr, val)

            else:
                val = fun(*args, **kwargs)

            return val
        return new_fun
    return decorate

class ElastodynamicsBaseTS(TimeSteppingSolver):
    """
    Base class for elastodynamics solvers.

    Assumes block-diagonal matrix in `u`, `v`, `a`.
    """

    _common_parameters = [
        ('t0', 'float', 0.0, False,
         'The initial time.'),
        ('t1', 'float', 1.0, False,
         'The final time.'),
        ('dt', 'float', None, False,
         'The time step. Used if `n_step` is not given.'),
        ('n_step', 'int', 10, False,
         'The number of time steps. Has precedence over `dt`.'),
        ('is_linear', 'bool', False, False,
         'If True, the problem is considered to be linear.'),
        ('has_time_derivatives', 'bool', False, False,
         """If True, the problem equations contain time derivatives of other
            variables besides displacements. In that case the cached constant
            matrices must be cleared on time step changes."""),
        ('var_names', 'dict', None, False,
         """The mapping of variables with keys 'u', 'du', 'ddu' and 'extra',
            and values corresponding to the names of the actual variables.
            See `var_names` returned from :func:`transform_equations_ed()`"""),
    ]

    def __init__(self, conf, nls=None, tsc=None, context=None, **kwargs):
        TimeSteppingSolver.__init__(self, conf, nls=nls, tsc=tsc,
                                    context=context, **kwargs)
        self.conf.quasistatic = False

        if self.tsc is None:
            self.tsc = FixedTSC({})

        if isinstance(self.tsc, FixedTSC):
            # Using TimeStepper instead of VariableTimeStepper ensures the
            # final time is reached "exactly".
            self.ts = TimeStepper.from_conf(self.conf)

        else:
            self.ts = VariableTimeStepper.from_conf(self.conf)

        nd = self.ts.n_digit
        format = '====== time %%e (step %%%dd of %%%dd) =====' % (nd, nd)

        self.format = format
        self.verbose = self.conf.verbose
        self.constant_matrices = None
        self.matrix = None

    def get_matrices(self, nls, vec, unpack=None):
        if self.conf.is_linear and self.constant_matrices is not None:
            out = self.constant_matrices

        else:
            aux = nls.fun_grad(vec)

            if unpack is not None:
                iue, iu, ie, iv, ia = unpack.indices
                aux = nls.fun_grad(vec)

                M = aux[ia, ia]
                C = aux[iv, iv]
                K = aux[iue, iue]

            else:
                assert_((len(vec) % 3) == 0)
                i3 = len(vec) // 3

                K = aux[:i3, :i3]
                C = aux[i3:2*i3, i3:2*i3]
                M = aux[2*i3:, 2*i3:]

            out = (M, C, K)

            if self.conf.is_linear:
                M.eliminate_zeros()
                C.eliminate_zeros()
                K.eliminate_zeros()
                self.constant_matrices = (M, C, K)

        return out

    def get_a0(self, nls, u0, e0, v0, unpack):
        iue, iu, ie, iv, ia = unpack.indices

        vec = nm.r_[u0, e0, v0, nm.zeros_like(v0)]
        aux = nls.fun(vec)
        r = aux[iu] + aux[iv] + aux[ia]

        M = self.get_matrices(nls, vec, unpack)[0][iu, iu]
        a0 = nls.lin_solver(-r, mtx=M)
        nls.lin_solver.clear()
        output_array_stats(a0, 'initial acceleration', verbose=self.verbose)
        return a0

    def get_initial_vec(self, nls, vec0, init_fun, prestep_fun, poststep_fun):
        if not set(self.conf.var_names.keys()).issuperset(['ddu', 'du', 'u']):
            raise ValueError(
                "var_names have to contain 'u', 'du', 'ddu' keys!"
            )

        unpack, pack = gen_multi_vec_packing(
            self.di, self.conf.var_names,
            extra_variables=self.get('extra_variables', False),
        )
        self.unpack = unpack
        self.pack = pack

        ts = self.ts
        vec0 = init_fun(ts, vec0)

        output(self.format % (ts.time, ts.step + 1, ts.n_step),
               verbose=self.verbose)
        if ts.step == 0:
            vec0 = prestep_fun(ts, vec0)
            if unpack.n_arg == 4:
                u0, e0, v0, _ = unpack(vec0)

            else:
                u0, v0, _ = unpack(vec0)
                e0 = nm.empty(0, dtype=u0.dtype)

            a0 = self.get_a0(nls, u0, e0, v0, unpack)

            if unpack.n_arg == 4:
                vec = pack(u0, e0, v0, a0)

            else:
                vec = pack(u0, v0, a0)

            vec = poststep_fun(ts, vec)
            ts.advance()

        else:
            vec = vec0

        return vec, unpack, pack

    def clear_lin_solver(self, clear_constant_matrices=True):
        self.nls.lin_solver.clear()
        self.matrix = None
        if clear_constant_matrices:
            self.constant_matrices = None

    @standard_ts_call
    def __call__(self, vec0=None, nls=None, init_fun=None, prestep_fun=None,
                 poststep_fun=None, status=None, **kwargs):
        sig = signature(init_fun)
        if len(sig.parameters) == 3:
            init_fun = partial(init_fun, self)

        sig = signature(poststep_fun)
        if len(sig.parameters) == 3:
            poststep_fun = partial(poststep_fun, self)

        vec, unpack, pack = self.get_initial_vec(
            nls, vec0, init_fun, prestep_fun, poststep_fun)

        ts = self.ts
        dt0 = self.tsc.get_initial_dt(ts, vec, unpack=unpack)
        if not isinstance(self.tsc, FixedTSC):
            ts.set_time_step(dt0, update_time=True)
        while 1:
            output(self.format % (ts.time, ts.step + 1, ts.n_step),
                   verbose=self.verbose)
            # step, time = time step to compute = n+1
            # step-1, time-ts.dt = previous known step data = n
            # adaptivity modifies dt and time.
            while 1:
                # Previous step state q(t_n).
                # Both prestep_fun() and poststep_fun() apply EBCs to vec/vect
                # in case active_only is False.
                # TODO: Generalized alpha: EBCs for current time t_{n+1}. but
                # loads should be applied in the mid-step time t_{n+1-a}.
                vec = prestep_fun(ts, vec)
                vect = self.step(ts, vec, nls, pack, unpack,
                                 prestep_fun=prestep_fun)

                if isinstance(self.tsc, FixedTSC):
                    new_dt = ts.dt
                    break

                new_dt, status = self.tsc(ts, vec, vect, unpack=unpack)
                output('dt:', ts.dt, 'new dt:', new_dt, 'status:', status,
                       verbose=self.verbose)
                if new_dt != ts.dt:
                    self.clear_lin_solver(
                        clear_constant_matrices=self.conf.has_time_derivatives,
                    )

                if status.result == 'accept':
                    break

                ts.set_time_step(new_dt, update_time=True)

            # Current step state q(t_{n+1}).
            vect = poststep_fun(ts, vect)

            if ts.nt >= 1:
                break

            if new_dt != ts.dt:
                ts.set_time_step(new_dt, update_time=False)
            ts.advance()

            vec = vect

        return vec

class VelocityVerletTS(ElastodynamicsBaseTS):
    """
    Solve elastodynamics problems by the explicit velocity-Verlet method.

    The algorithm can be found in [1]_.

    It is mathematically equivalent to the :class:`CentralDifferenceTS` method.
    The current implementation code is essentially the same, as the mid-time
    velocities are not used for anything other than computing the new time
    velocities.

    .. [1] Swope, William C.; H. C. Andersen; P. H. Berens; K. R. Wilson (1
           January 1982). "A computer simulation method for the calculation of
           equilibrium constants for the formation of physical clusters of
           molecules: Application to small water clusters". The Journal of
           Chemical Physics. 76 (1): 648 (Appendix). doi:10.1063/1.442716
    """
    name = 'ts.velocity_verlet'

    _parameters = [
    ] + ElastodynamicsBaseTS._common_parameters

    def create_nlst(self, nls, dt, u0, v0, a0):
        vm = v0 + 0.5 * dt * a0
        u1 = u0 + dt * vm
        def v1(a):
            return vm + 0.5 * dt * a

        nlst = nls.copy()

        def fun(at):
            vec = nm.r_[u1, vm, at]

            aux = nls.fun(vec)

            i3 = len(at)
            rt = aux[:i3] + aux[i3:2*i3] + aux[2*i3:]
            return rt

        @_cache(self, 'matrix', self.conf.is_linear)
        def fun_grad(at):
            vec = None if self.conf.is_linear else nm.r_[u1, vm, at]
            M = self.get_matrices(nls, vec)[0]
            return M

        nlst.fun = fun
        nlst.fun_grad = fun_grad
        nlst.v1 = v1
        nlst.u1 = u1

        return nlst

    def step(self, ts, vec, nls, pack, unpack, **kwargs):
        """
        Solve a single time step.
        """
        dt = ts.dt
        ut, vt, at = unpack(vec)

        nlst = self.create_nlst(nls, dt, ut, vt, at)
        atp = nlst(at)
        vtp = nlst.v1(atp)
        utp = nlst.u1

        vect = pack(utp, vtp, atp)

        return vect

class CentralDifferenceTS(ElastodynamicsBaseTS):
    r"""
    Solve elastodynamics problems by the explicit central difference method.

    It is the same method as obtained by using :class:`NewmarkTS` with
    :math:`\beta = 0`, :math:`\gamma = 1/2`, but uses a simpler code.

    It is also mathematically equivalent to the :class:`VelocityVerletTS`
    method. The current implementation code is essentially the same.

    This solver supports, when used with :class:`RMMSolver
    <sfepy.solvers.ls.RMMSolver>`, the reciprocal mass matrix algorithm, see
    :class:`MassTerm <sfepy.terms.terms_mass.MassTerm>`.
    """
    name = 'ts.central_difference'

    _parameters = [
    ] + ElastodynamicsBaseTS._common_parameters

    def _create_nlst_a(self, nls, dt, ufun, vfun, cc, cache_name, is_rmm=False):
        nlst = nls.copy()

        if is_rmm:
            def fun(at):
                ut = ufun()
                zz = nm.zeros_like(ut)
                vec = nm.r_[ut, zz, zz]

                aux = nls.fun(vec)

                i3 = len(at)
                rt = aux[:i3]
                return rt

            @_cache(self, cache_name, self.conf.is_linear)
            def fun_grad(at):
                M = self.get_matrices(nls, None)[0]

                return M

        else:
            def fun(at):
                vec = nm.r_[ufun(), vfun(at), at]

                aux = nls.fun(vec)

                i3 = len(at)
                rt = aux[:i3] + aux[i3:2*i3] + aux[2*i3:]
                return rt

            @_cache(self, cache_name, self.conf.is_linear)
            def fun_grad(at):
                vec = (None if self.conf.is_linear
                       else nm.r_[ufun(), vfun(at), at])
                M, C = self.get_matrices(nls, vec)[:2]

                Kt = M + cc * C
                return Kt

        nlst.fun = fun
        nlst.fun_grad = fun_grad
        nlst.u = ufun
        nlst.v = vfun

        return nlst

    def create_nlst(self, nls, dt, u0, v0, a0):
        dt2 = dt**2

        def v(a):
            return v0 + dt * 0.5 * (a0 + a)

        def u():
            return u0 + dt * v0 + dt2 * 0.5 * a0

        if isinstance(nls.lin_solver, RMMSolver):
            import scipy.sparse as sps
            class NoNLS(NonlinearSolver):

                def __call__(self, vec_x0, conf=None, fun=None, fun_grad=None,
                             lin_solver=None, iter_hook=None, status=None):
                    vec_r = self.fun(vec_x0)
                    # Dummy all-zero matrix to make standard_call() happy.
                    mtx_a = sps.csr_matrix((vec_r.shape[0], vec_r.shape[0]))
                    return self.lin_solver(-vec_r, mtx=mtx_a)

            nlst = self._create_nlst_a(nls, dt, u, v, 0.5 * dt, 'matrix',
                                       is_rmm=True)
            nlst = NoNLS(Struct(name='nonls', kind='nls.nonls'),
                         fun=nlst.fun, fun_grad=nlst.fun_grad,
                         lin_solver=nlst.lin_solver, iter_hook=nlst.iter_hook,
                         status=nlst.status, context=nlst.context,
                         u=nlst.u, v=nlst.v)
            # nlst.lin_solver.a0 = a0

        else:
            nlst = self._create_nlst_a(nls, dt, u, v, 0.5 * dt, 'matrix')

        return nlst

    def step(self, ts, vec, nls, pack, unpack, **kwargs):
        """
        Solve a single time step.
        """
        dt = ts.dt
        ut, vt, at = unpack(vec)

        nlst = self.create_nlst(nls, dt, ut, vt, at)
        atp = nlst(at)
        vtp = nlst.v(atp)
        utp = nlst.u()

        vect = pack(utp, vtp, atp)

        return vect

class NewmarkTS(ElastodynamicsBaseTS):
    r"""
    Solve elastodynamics problems by the Newmark method.

    The method was introduced in [1]. Common settings [2]:

    ==================== ======== ==== ===== ==========
    name                 kind     beta gamma Omega_crit
    ==================== ======== ==== ===== ==========
    trapezoidal rule:    implicit 1/4  1/2   unconditional
    linear acceleration: implicit 1/6  1/2   :math:`2\sqrt{3}`
    Fox-Goodwin:         implicit 1/12 1/2   :math:`\sqrt{6}`
    central difference:  explicit 0    1/2   2
    ==================== ======== ==== ===== ==========

    All of these methods are 2-order of accuracy.

    [1] Newmark, N. M. (1959) A method of computation for structural dynamics.
    Journal of Engineering Mechanics, ASCE, 85 (EM3) 67-94.

    [2] Arnaud Delaplace, David Ryckelynck: Solvers for Computational Mechanics
    """
    name = 'ts.newmark'
    extra_variables = True

    _parameters = [
        ('beta', 'float', 0.25, False, 'The Newmark method parameter beta.'),
        ('gamma', 'float', 0.5, False, 'The Newmark method parameter gamma.'),
    ] + ElastodynamicsBaseTS._common_parameters

    def create_nlst(self, nls, dt, gamma, beta, u0, e0, v0, a0, pack, unpack):
        dt2 = dt**2
        iue, iu, ie, iv, ia = pack.indices

        cc0 = (1.0 - gamma) * dt
        cc = gamma * dt
        ck0 = (0.5 - beta) * dt2
        ck = beta * dt2

        def v(a):
            return v0 + cc0 * a0 + cc * a

        def u(a):
            return u0 + dt * v0 + ck0 * a0 + ck * a

        if iue == iu:
            def fun(at):
                vec = nm.r_[u(at), v(at), at]

                aux = nls.fun(vec)

                rt = aux[iu] + aux[iv] + aux[ia]
                return rt

            @_cache(self, 'matrix', self.conf.is_linear)
            def fun_grad(at):
                vec = None if self.conf.is_linear else nm.r_[u(at), v(at), at]
                M, C, K = self.get_matrices(nls, vec, unpack)

                Kt = M + cc * C + ck * K
                return Kt

        else: # Extra variables present.
            def fun(aet):
                at = aet[iu]
                vec = nm.r_[u(at), aet[ie], v(at), at]

                aux = nls.fun(vec)

                rt = nm.empty(pack.n_uedof, aux.dtype)
                rt[iu] = aux[iu] + aux[iv] + aux[ia]
                rt[ie] = aux[ie]
                return rt

            @_cache(self, 'matrix', self.conf.is_linear)
            def fun_grad(aet):
                if self.conf.is_linear:
                    vec = None

                else:
                    at = aet[iu]
                    vec = nm.r_[u(at), aet[ie], v(at), at]

                M, C, K = self.get_matrices(nls, vec, unpack)

                Kt = K.copy()
                Kt[:, iu] *= ck
                Kt[iu, iu] += M + cc * C
                return Kt

        nlst = nls.copy()
        nlst.fun = fun
        nlst.fun_grad = fun_grad
        nlst.u = u
        nlst.v = v

        return nlst

    def step(self, ts, vec, nls, pack, unpack, **kwargs):
        """
        Solve a single time step.
        """
        dt = ts.dt
        conf = self.conf
        ut, et, vt, at = unpack(vec)
        nlst = self.create_nlst(nls, dt, conf.gamma, conf.beta, ut, et, vt, at,
                                pack, unpack)

        aetp = nlst(pack(at, et))
        atp, etp = unpack(aetp, mode='nls')
        vtp = nlst.v(atp)
        utp = nlst.u(atp)

        vect = pack(utp, etp, vtp, atp)
        return vect

class GeneralizedAlphaTS(ElastodynamicsBaseTS):
    r"""
    Solve elastodynamics problems by the generalized :math:`\alpha` method.

    - The method was introduced in [1].
    - The method is unconditionally stable provided :math:`\alpha_m \leq
      \alpha_f \leq \frac{1}{2}`, :math:`\beta >= \frac{1}{4} +
      \frac{1}{2}(\alpha_f - \alpha_m)`.
    - The method is second-order accurate provided :math:`\gamma = \frac{1}{2} -
      \alpha_m + \alpha_f`. This is used when `gamma` is ``None``.
    - High frequency dissipation is maximized for :math:`\beta = \frac{1}{4}(1
      - \alpha_m + \alpha_f)^2`. This is used when `beta` is ``None``.
    - The default values of :math:`\alpha_m`, :math:`\alpha_f` (if `alpha_m` or
      `alpha_f`  are ``None``) are based on the user specified high-frequency
      dissipation parameter `rho_inf`.

    Special settings:

    - :math:`\alpha_m = 0` corresponds to the HHT-:math:`\alpha` method.
    - :math:`\alpha_f = 0` corresponds to the WBZ-:math:`\alpha` method.
    - :math:`\alpha_m = 0`, :math:`\alpha_f = 0` produces the Newmark method.

    [1] J. Chung, G.M.Hubert. "A Time Integration Algorithm for Structural
    Dynamics with Improved Numerical Dissipation: The
    Generalized-:math:`\alpha` Method" ASME Journal of Applied Mechanics, 60,
    371:375, 1993.
    """
    name = 'ts.generalized_alpha'

    _parameters = [
        ('rho_inf', 'float', 0.5, False,
         """The spectral radius in the high frequency limit (user specified
            high-frequency dissipation) in [0, 1]:
            1 = no dissipation, 0 = asymptotic annihilation."""),
        ('alpha_m', 'float', None, False,
         r'The parameter :math:`\alpha_m`.'),
        ('alpha_f', 'float', None, False,
         r'The parameter :math:`\alpha_f`.'),
        ('beta', 'float', None, False,
         r'The Newmark-like parameter :math:`\beta`.'),
        ('gamma', 'float', None, False,
         r'The Newmark-like parameter :math:`\gamma`.'),
    ] + ElastodynamicsBaseTS._common_parameters

    def __init__(self, conf, nls=None, tsc=None, context=None, **kwargs):
        ElastodynamicsBaseTS.__init__(self, conf, nls=nls, tsc=tsc,
                                      context=context, **kwargs)
        conf = self.conf

        rho_inf = conf.rho_inf
        alpha_m = get_default(conf.alpha_m,
                              (2.0 * rho_inf - 1.0) / (rho_inf + 1.0))
        alpha_f = get_default(conf.alpha_f, rho_inf / (rho_inf + 1.0))
        beta = get_default(conf.beta, 0.25 * (1.0 - alpha_m + alpha_f)**2)
        gamma = get_default(conf.gamma, 0.5 - alpha_m + alpha_f)
        self.pars = (alpha_m, alpha_f, gamma, beta)

        output('parameters rho_inf, alpha_m, alpha_f, beta, gamma:',
               verbose=self.verbose)
        output(rho_inf, alpha_m, alpha_f, beta, gamma,
               verbose=self.verbose)

    def _create_nlst_a(self, nls, dt, ufun, vfun, afun, cm, cc, ck, cache_name):
        nlst = nls.copy()

        def fun(at):
            vec = nm.r_[ufun(at), vfun(at), afun(at)]

            aux = nls.fun(vec)

            i3 = len(at)
            rt = aux[:i3] + aux[i3:2*i3] + aux[2*i3:]
            return rt

        @_cache(self, cache_name, self.conf.is_linear)
        def fun_grad(at):
            vec = (None if self.conf.is_linear
                   else nm.r_[ufun(at), vfun(at), afun(at)])
            M, C, K = self.get_matrices(nls, vec)

            Kt = cm * M + cc * C + ck * K
            return Kt

        nlst.fun = fun
        nlst.fun_grad = fun_grad

        return nlst

    def create_nlst(self, nls, dt, alpha_m, alpha_f, gamma, beta, u0, v0, a0):
        dt2 = dt**2

        def u1(a):
            return u0 + dt * v0 + dt2 * ((0.5 - beta) * a0 + beta * a)

        def v1(a):
            return v0 + dt * ((1.0 - gamma) * a0 + gamma * a)

        def um(a):
            return (1.0 - alpha_f) * u1(a) + alpha_f * u0

        def vm(a):
            return (1.0 - alpha_f) * v1(a) + alpha_f * v0

        def am(a):
            return (1.0 - alpha_m) * a + alpha_m * a0

        nlst = self._create_nlst_a(nls, dt, um, vm, am,
                                   (1.0 - alpha_m),
                                   (1.0 - alpha_f) * gamma * dt,
                                   (1.0 - alpha_f) * beta * dt2,
                                   'matrix')
        nlst.u = u1
        nlst.v = v1

        return nlst

    def step(self, ts, vec, nls, pack, unpack, **kwargs):
        """
        Solve a single time step.
        """
        dt = ts.dt
        alpha_m, alpha_f, gamma, beta = self.pars
        # Previous step state q(t_n).
        ut, vt, at = unpack(vec)

        nlst = self.create_nlst(nls, dt, alpha_m, alpha_f, gamma, beta,
                                ut, vt, at)

        # Notation: a = \alpha_f, t = t_{n+1}, t - dt = t_n.
        # Set time to t_{n+1-a} = (1 - a) t + a (t - dt) = t - a dt
        ts.set_substep_time(- alpha_f * dt)
        atp = nlst(at)
        # Restore t_{n+1}.
        ts.restore_step_time()

        vtp = nlst.v(atp)
        utp = nlst.u(atp)
        # Current step state q(t_{n+1}).
        vect = pack(utp, vtp, atp)

        return vect

class BatheTS(ElastodynamicsBaseTS):
    """
    Solve elastodynamics problems by the Bathe method.

    The method was introduced in [1].

    [1] Klaus-Juergen Bathe, Conserving energy and momentum in nonlinear
    dynamics: A simple implicit time integration scheme, Computers &
    Structures, Volume 85, Issues 7-8, 2007, Pages 437-445, ISSN 0045-7949,
    https://doi.org/10.1016/j.compstruc.2006.09.004.
    """
    name = 'ts.bathe'

    _parameters = [
    ] + ElastodynamicsBaseTS._common_parameters

    def __init__(self, conf, nls=None, context=None, **kwargs):
        ElastodynamicsBaseTS.__init__(self, conf, nls=nls, context=context,
                                      **kwargs)
        self.matrix1 = None
        self.ls1 = None
        self.ls2 = None

    def _create_nlst_u(self, nls, dt, vfun, afun, cm, cc, cache_name):
        nlst = nls.copy()

        def fun(ut):
            vt = vfun(ut)
            at = afun(vt)
            vec = nm.r_[ut, vt, at]

            aux = nls.fun(vec)

            i3 = len(at)
            rt = aux[:i3] + aux[i3:2*i3] + aux[2*i3:]
            return rt

        @_cache(self, cache_name, self.conf.is_linear)
        def fun_grad(ut):
            if self.conf.is_linear:
                vec = None
            else:
                vt = vfun(ut)
                at = afun(vt)
                vec = nm.r_[ut, vt, at]

            M, C, K = self.get_matrices(nls, vec)

            Kt = cm * M + cc * C + K
            return Kt

        nlst.fun = fun
        nlst.fun_grad = fun_grad
        nlst.v = vfun
        nlst.a = afun

        return nlst

    def create_nlst1(self, nls, dt, u0, v0, a0):
        """
        The first sub-step: the trapezoidal rule.
        """
        dt4 = 4.0 / dt

        def v(u):
            return dt4 * (u - u0) - v0

        def a(v):
            return dt4 * (v - v0) - a0

        nlst = self._create_nlst_u(nls, dt, v, a, dt4 * dt4, dt4, 'matrix1')
        if self.ls1 is None:
            self.ls1 = nls.lin_solver.copy()

        nlst.lin_solver = self.ls1
        return nlst

    def create_nlst2(self, nls, dt, u0, u1, v0, v1):
        """
        The second sub-step: the three-point Euler backward method.
        """
        dt1 = 1.0 / dt
        dt4 = 4.0 * dt1
        dt3 = 3.0 * dt1

        def v(u):
            return dt1 * u0 - dt4 * u1 + dt3 * u

        def a(v):
            return dt1 * v0 - dt4 * v1 + dt3 * v

        nlst = self._create_nlst_u(nls, dt, v, a, dt3 * dt3, dt3, 'matrix')
        if self.ls2 is None:
            self.ls2 = nls.lin_solver.copy()

        nlst.lin_solver = self.ls2
        return nlst

    def clear_lin_solver(self, clear_constant_matrices=True):
        ElastodynamicsBaseTS.clear_lin_solver(
            self, clear_constant_matrices=clear_constant_matrices,
        )
        self.ls1 = self.ls2 = None
        self.matrix1 = None

    def step(self, ts, vec, nls, pack, unpack, prestep_fun):
        """
        Solve a single time step.
        """
        dt = ts.dt
        ut, vt, at = unpack(vec)
        nlst1 = self.create_nlst1(nls, dt, ut, vt, at)
        ut1 = nlst1(ut)
        vt1 = nlst1.v(ut1)
        at1 = nlst1.a(vt1)

        ts.set_substep_time(0.5 * dt)

        vec1 = pack(ut1, vt1, at1)
        vec1 = prestep_fun(ts, vec1)

        nlst2 = self.create_nlst2(nls, dt, ut, ut1, vt, vt1)
        ut2 = nlst2(ut1)
        vt2 = nlst2.v(ut2)
        at2 = nlst2.a(vt2)

        ts.restore_step_time()

        vec2 = pack(ut2, vt2, at2)

        return vec2



import pytest

from sfepy import data_dir
import sfepy.base.testing as tst

filename_mesh = data_dir + '/meshes/2d/square_unit_tri.mesh'

material_1 = {
    'name' : 'coef',
    'values' : {
        'val' : 1.0,
    },
}
material_2 = {
    'name' : 'm',
    'values' : {
        'K' : [[1.0, 0.0], [0.0, 1.0]],
    },
}

field_1 = {
    'name' : 'a_harmonic_field',
    'dtype' : 'real',
    'shape' : 'scalar',
    'region' : 'Omega',
    'approx_order' : 2,
}

variable_1 = {
    'name' : 't',
    'kind' : 'unknown field',
    'field' : 'a_harmonic_field',
    'order' : 0,
}
variable_2 = {
    'name' : 's',
    'kind' : 'test field',
    'field' : 'a_harmonic_field',
    'dual' : 't',
}

region_1000 = {
    'name' : 'Omega',
    'select' : 'all',
}

region_1 = {
    'name' : 'Left',
    'select' : 'vertices in (x < -0.499)',
    'kind' : 'facet',
}
region_2 = {
    'name' : 'Right',
    'select' : 'vertices in (x > 0.499)',
    'kind' : 'facet',
}
region_3 = {
    'name' : 'Gamma',
    'select' : 'vertices of surface',
    'kind' : 'facet',
}

ebc_1 = {
    'name' : 't_left',
    'region' : 'Left',
    'dofs' : {'t.0' : 5.0},
}
ebc_2 = {
    'name' : 't_right',
    'region' : 'Right',
    'dofs' : {'t.0' : 0.0},
}

integral_1 = {
    'name' : 'i',
    'order' : 2,
}

equations = {
    'Temperature' : """dw_laplace.i.Omega(coef.val, s, t) = 0"""
}

solution = {
    't' : '- 5.0 * (x - 0.5)',
}

solver_0 = {
    'name' : 'ls',
    'kind' : 'ls.scipy_direct',
}

solver_1 = {
    'name' : 'newton',
    'kind' : 'nls.newton',

    'i_max'      : 1,
    'eps_a'      : 1e-10,
}

lin_min, lin_max = 0.0, 2.0

def linear(bc, ts, coor, which):
    vals = coor[:,which]
    min_val, max_val = vals.min(), vals.max()
    vals = (vals - min_val) / (max_val - min_val) * (lin_max - lin_min) + lin_min
    return vals

def linear_x(bc, ts, coor):
    return linear(bc, ts, coor, 0)
def linear_y(bc, ts, coor):
    return linear(bc, ts, coor, 1)
def linear_z(bc, ts, coor):
    return linear(bc, ts, coor, 2)

@pytest.fixture(scope='module')
def data():
    import sys
    from sfepy.applications import solve_pde
    from sfepy.base.conf import ProblemConf
    from sfepy.base.base import Struct

    conf = ProblemConf.from_dict(globals(), sys.modules[__name__])
    problem, state = solve_pde(conf, save_results=False)

    return Struct(problem=problem, state=state)

def test_solution(data):
    problem = data.problem
    sol = problem.conf.solution
    vec = data.state()

    variables = problem.get_variables()

    ok = True
    for var_name, expression in sol.items():
        coor = variables[var_name].field.get_coor()
        ana_sol = tst.eval_coor_expression(expression, coor)
        num_sol = variables.get_vec_part(vec, var_name)
        ret = tst.compare_vectors(ana_sol, num_sol,
                                  label1='analytical %s' % var_name,
                                  label2='numerical %s' % var_name)
        if not ret:
            tst.report('variable %s: failed' % var_name)

        ok = ok and ret

    assert ok

def test_boundary_fluxes(data, output_dir):
    import os.path as op
    from sfepy.linalg import rotation_matrix2d
    from sfepy.discrete import Material
    problem = data.problem

    angles = [0, 30, 45]
    region_names = ['Left', 'Right', 'Gamma']
    values = [5.0, -5.0, 0.0]

    variables = problem.get_variables()
    get_state = variables.get_vec_part
    state = data.state.copy()

    problem.time_update(ebcs={}, epbcs={})

    state.apply_ebc()
    nls = problem.get_nls()
    aux = nls.fun(state())

    field = variables['t'].field

    conf_m = problem.conf.get_item_by_name('materials', 'm')
    m = Material.from_conf(conf_m, problem.functions)

    name = op.join(output_dir,
                   op.split(problem.domain.mesh.name)[1] + '_%02d.mesh')

    orig_coors = problem.get_mesh_coors().copy()
    ok = True
    for ia, angle in enumerate(angles):
        tst.report('%d: mesh rotation %d degrees' % (ia, angle))
        problem.domain.mesh.transform_coors(rotation_matrix2d(angle),
                                             ref_coors = orig_coors)
        problem.set_mesh_coors(problem.domain.mesh.coors,
                               update_fields=True)
        problem.domain.mesh.write(name % angle, io='auto')
        for ii, region_name in enumerate(region_names):
            flux_term = 'ev_surface_flux.i.%s(m.K, t)' % region_name
            val1 = problem.evaluate(flux_term, t=variables['t'], m=m)

            rvec = get_state(aux, 't', True)
            reg = problem.domain.regions[region_name]
            nods = field.get_dofs_in_region(reg, merge=True)
            val2 = rvec[nods].sum() # Assume 1 dof per node.

            ok = ok and ((abs(val1 - values[ii]) < 1e-10) and
                         (abs(val2 - values[ii]) < 1e-10))
            tst.report('  %d. %s: %e == %e == %e'
                       % (ii, region_name, val1, val2, values[ii]))

    # Restore original coordinates.
    problem.domain.mesh.transform_coors(rotation_matrix2d(0),
                                        ref_coors=orig_coors)
    problem.set_mesh_coors(problem.domain.mesh.coors,
                           update_fields=True)

    assert ok



"""
Test computing volumes by volume or surface integrals.
"""
import numpy as nm
import pytest

from sfepy import data_dir
import sfepy.base.testing as tst

filename_mesh = data_dir + '/meshes/3d/elbow.mesh'

fields = {
    'scalar' : ('real', 'scalar', 'Omega', 1),
    'vector' : ('real', 'vector', 'Omega', 1),
}

integrals = {
    'i' : 2,
}

regions = {
    'Omega' : 'all',
    'Gamma' : ('vertices of surface', 'facet'),
}

expressions = {
    'volume_p' : 'ev_volume.i.Omega(p)',
    'volume_u' : 'ev_volume.i.Omega(u)',
    'surface_p' : 'ev_volume_surface.i.Gamma(p)',
    'surface_u' : 'ev_volume_surface.i.Gamma(u)',
}

@pytest.fixture(scope='module')
def problem():
    import sys
    from sfepy.discrete import Problem
    from sfepy.base.conf import ProblemConf

    conf = ProblemConf.from_dict(globals(), sys.modules[__name__])

    problem = Problem.from_conf(conf, init_equations=False)
    return problem

def test_volume(problem):
    from sfepy.discrete import FieldVariable

    ok = True

    field_map = {'u' : 'vector', 'p' : 'scalar'}

    volumes = {}
    avg = 0.0
    for key, term in expressions.items():
        var_name = key[-1]
        field = problem.fields[field_map[var_name]]
        var = FieldVariable(var_name, 'parameter', field,
                            primary_var_name='(set-to-None)')

        val = problem.evaluate(term, **{var_name : var})

        volumes[key] = val
        avg += val

    avg /= len(volumes)

    for key, val in volumes.items():
        err = nm.abs(avg - val) / nm.abs(avg)
        _ok = err < 1e-12
        tst.report('"'"%s"'" - volume: %e' % (key, val))
        tst.report('"'"%s"'" - relative volume difference: %e -> %s'
                    % (key, err, _ok))
        ok = ok and _ok

    assert ok

def test_volume_tl(problem):
    from sfepy.discrete import FieldVariable

    fu = problem.fields['vector']
    fq = problem.fields['scalar']

    var_u = FieldVariable('u', 'parameter', fu,
                          primary_var_name='(set-to-None)')
    var_q = FieldVariable('q', 'test', fq,
                          primary_var_name='(set-to-None)')

    var_u.set_data(nm.linspace(0, 0.004, var_u.n_dof))

    vval = problem.evaluate('dw_tl_volume.i.Omega( q, u )',
                            term_mode='volume', q=var_q, u=var_u)

    sval = problem.evaluate('ev_tl_volume_surface.i.Gamma( u )',
                            u=var_u)

    ok = abs(vval - sval) < 1e-14

    tst.report('TL: by volume: %e == by surface: %e -> %s' %
                (vval, sval, ok))

    assert ok



import pytest

from sfepy import data_dir
import sfepy.base.testing as tst

filename_mesh = data_dir + '/meshes/2d/circle_sym.mesh'

material_1 = {
    'name' : 'coef',
    'values' : {
        'val' : 1.0,
    },
}
material_2 = {
    'name' : 'm',
    'values' : {
        'K' : [[1.0, 0.0], [0.0, 1.0]],
    },
}

field_1 = {
    'name' : 'a_harmonic_field',
    'dtype' : 'real',
    'shape' : 'scalar',
    'region' : 'Omega',
    'approx_order' : 2,
}

variable_1 = {
    'name' : 't',
    'kind' : 'unknown field',
    'field' : 'a_harmonic_field',
    'order' : 0,
}
variable_2 = {
    'name' : 's',
    'kind' : 'test field',
    'field' : 'a_harmonic_field',
    'dual' : 't',
}

region_1000 = {
    'name' : 'Omega',
    'select' : 'all',
}

region_1 = {
    'name' : 'Centre',
    'select' : 'vertices in (x < 1e-8) & (x > -1e-8) & (y < 1e-8) & (y > -1e-8)',
    'kind' : 'vertex'
}

region_2 = {
    'name' : 'Gamma',
    'select' : 'vertices of surface',
    'kind' : 'facet',
}

ebc_1 = {
    'name' : 't_centre',
    'region' : 'Centre',
    'dofs' : {'t.0' : 1.0},
}
ebc_2 = {
    'name' : 't_gamma',
    'region' : 'Gamma',
    'dofs' : {'t.0' : 0.0},
}

integral_1 = {
    'name' : 'i',
    'order' : 2,
}

equations = {
    'Temperature' : """dw_laplace.i.Omega(coef.val, s, t) = 0"""
}

solution = {
    't' : '- 5.0 * (x - 0.5)',
}

solver_0 = {
    'name' : 'ls',
    'kind' : 'ls.scipy_direct',
}

solver_1 = {
    'name' : 'newton',
    'kind' : 'nls.newton',

    'i_max'      : 1,
    'eps_a'      : 1e-10,
}

@pytest.fixture(scope='module')
def data():
    import sys
    from sfepy.applications import solve_pde
    from sfepy.base.conf import ProblemConf
    from sfepy.base.base import Struct

    conf = ProblemConf.from_dict(globals(), sys.modules[__name__])
    problem, state = solve_pde(conf, save_results=False)

    return Struct(problem=problem, state=state)

def test_boundary_fluxes(data):
    from sfepy.discrete import Material
    problem = data.problem

    region_names = ['Gamma']

    variables = problem.get_variables()
    get_state = variables.get_vec_part
    state = data.state.copy()

    problem.time_update(ebcs={}, epbcs={})

    state.apply_ebc()
    nls = problem.get_nls()
    aux = nls.fun(state())

    field = variables['t'].field

    conf_m = problem.conf.get_item_by_name('materials', 'm')
    m = Material.from_conf(conf_m, problem.functions)

    ok = True
    for ii, region_name in enumerate(region_names):
        flux_term = 'ev_surface_flux.1.%s(m.K, t)' % region_name
        val1 = problem.evaluate(flux_term, t=variables['t'], m=m)

        rvec = get_state(aux, 't', True)
        reg = problem.domain.regions[region_name]
        nods = field.get_dofs_in_region(reg, merge=True)
        val2 = rvec[nods].sum() # Assume 1 dof per node.

        eps = 1e-2
        ok = ok and ((abs(val1 - val2) < eps))
        tst.report('%d. %s: |%e - %e| = %e < %.2e'\
                   % (ii, region_name, val1, val2, abs(val1 - val2), eps))

    assert ok
